import uuid
from loguru import logger
import maya
import pandas as pd


from jamboree import JamboreeNew
from jamboree.handlers.complex.backtestable import BacktestDBHandler
from jamboree.handlers.processors import DynamicResample, DataProcessorsAbstract
from jamboree.utils.core import omit



class MetricHandler(BacktestDBHandler):
    """ 
        # METRIC HANDLER
        ---

        A simple metric handler. To be used with all machine learning related functions.

        Given an episode and other crucial information, we'll give all information regarding how something has progressed. 

        Some considerations:

        * We'll load the most recent metrics into the metadata handler (which will let us search for the metric)
        * We want to know how a batch model has been doing between episodes
            * Something that sends aggregation commands for all models that have been touched for a given category, subcategory, name set
            * That would also require us to know how to preprocess that information prior to adding it into the database
            * We'd also want to not do this too often to reduce CPU initialization load
            * A form or rate limiter
        

    """

    def __init__(self):
        super().__init__()
        self.entity = "metric"
        self.required = {
            "category": str,
            "subcategories": dict,
            "name": str
        }
        
        self._preprocessor: DataProcessorsAbstract = DynamicResample("data")
        

    @property
    def preprocessor(self) -> DataProcessorsAbstract:
        return self._preprocessor
    
    @preprocessor.setter
    def preprocessor(self, _preprocessor: DataProcessorsAbstract):
        self._preprocessor = _preprocessor

    def log(self, metric_dict:dict):
        """ Logs a metrics at the current time """
        current_time = self.time.head
        metric_dict['time'] = current_time
        metric_dict['timestamp'] = maya.now()._epoch
        # Add something here to make this searchable as well.
        self.save(metric_dict)
    
    def latest(self):
        """ Get the latest """
        _latest = self.last(ar='relative')
        omitted = omit(['episode', 'mtype', 'live', 'category', 'subcategories', 'type', 'name'], _latest)
        return omitted

    def reset_current_metric(self):
        self['episode'] = self.episode
        self['live'] = self.live

    def reset(self):
        """ Reset the data we're querying for. """
        super().reset()
        self.reset_current_metric()
    
    def step_time(self):
        """ """
        self.time.step()
        pass
    
    
    
    def __str__(self) -> str:
        name = self["name"]
        category = self["category"]
        subcategories = self["subcategories"]
        jscat = self.main_helper.generate_hash(subcategories)
        return f"{name}:{category}:{jscat}"

def metric_test():
    """ Test monitoring an online learning algorithm (using creme). """
    import random
    
    jambo = JamboreeNew()
    metric_log = MetricHandler()
    metric_log['category'] = "model"
    metric_log['subcategories'] = {}
    metric_log['name'] = "general_regressor"
    metric_log.processor = jambo
    metric_log.reset()
    metric_log.time.change_stepsize(hours=0, microseconds=10)
    while True:
        metric_log.reset_current_metric()
        metric_schema = {
            "accuracy": random.uniform(0, 1),
            "f1": random.uniform(0, 1)
        }
        metric_log.log(metric_schema)
        saved_metric = metric_log.latest()
        metric_log.step_time()
        print(saved_metric)

    
if __name__ == "__main__":
    metric_test()
    # import pandas_datareader.data as web
    # data_msft = web.DataReader('MSFT','yahoo',start='2010/1/1',end='2020/1/30').round(2)
    # data_apple = web.DataReader('AAPL','yahoo',start='2010/1/1',end='2020/1/30').round(2)
    # print(data_apple)
    # episode_id = uuid.uuid4().hex
    # jambo = Jamboree()
    # jam_processor = JamboreeNew()
    # data_hander = DataHandler()
    # data_hander.event = jambo
    # data_hander.processor = jam_processor
    # # The episode and live parameters are probably not good for the scenario. Will probably need to switch to something else to identify data
    # data_hander.episode = episode_id
    # data_hander.live = False
    # data_hander['category'] = "markets"
    # data_hander['subcategories'] = {
    #     "market": "stock",
    #     "country": "US",
    #     "sector": "techologyyyyyyyy"
    # }
    # data_hander['name'] = "MSFT"
    # data_hander.reset()
    # data_hander.store_time_df(data_msft, is_bar=True)


    # data_hander['name'] = "AAPL"
    # data_hander.store_time_df(data_apple, is_bar=True)
    # data_hander.reset()

    # data_hander.time.head = maya.now().subtract(weeks=200, hours=14)._epoch
    # data_hander.time.change_stepsize(microseconds=0, days=1, hours=0)
    # data_hander.time.change_lookback(microseconds=0, weeks=4, hours=0)

    
    # while data_hander.is_next:
    #     logger.info(magenta(data_hander.time.head, bold=True))
    #     print(data_hander.dataframe_from_head())
    #     data_hander.time.step()


import base64
import random
from abc import ABC
from copy import copy
from multiprocessing import cpu_count
from typing import List

import maya
import orjson
import ujson
from loguru import logger
from pebble.pool import ThreadPool
from redis import Redis

from funtime import Store
from jamboree.storage.databases import (MongoDatabaseConnection,
                                        RedisDatabaseConnection)


class EventProcessor(ABC):
    def save(self, query: dict, data: dict, abs_rel="absolute"):
        raise NotImplementedError


    def save_many(self, query: dict, data: List[dict], abs_rel="absolute"):
        raise NotImplementedError    


    def get_latest(self, query, abs_rel="absolute") -> dict:
        raise NotImplementedError


    def get_latest_many(self, query, abs_rel="absolute", limit=1000):
        raise NotImplementedError


    def get_between(self, query:dict, min_epoch:float, max_epoch:float, abs_rel:str="absolute") -> list:
        raise NotImplementedError


    def get_latest_by(self, query:dict, max_epoch, abs_rel="absolute", limit:int=10) -> dict:
        raise NotImplementedError


    def count(self, query: dict) -> int:
        raise NotImplementedError


    def remove_first(self, query: dict):
        raise NotImplementedError

    
    def pop_multiple(self, query: dict, limit: int):
        raise NotImplementedError


    def _bulk_save(self, query: dict, data: list):
        raise NotImplementedError


    def single_get(self, query:dict):
        raise NotImplementedError


    def single_set(self, query:dict, data:dict):
        raise NotImplementedError


    def single_delete(self, query:dict):
        raise NotImplementedError


class Jamboree(EventProcessor):
    """Adds and retrieves events at extremely fast speeds. Use to handle portfolio and trade information quickly."""

    def __init__(self, mongodb_host="localhost", redis_host="localhost", redis_port=6379):
        self.redis = Redis(redis_host, port=redis_port)
        self.store = Store(mongodb_host).create_lib('events').get_store()['events']
        self.pool = ThreadPool(max_workers=cpu_count() * 4)
        self.mongo_conn = MongoDatabaseConnection()
        self.redis_conn = RedisDatabaseConnection()
        self.mongo_conn.connection = self.store
        self.redis_conn.connection = self.redis
        # self.redis_conn.pool = self.pool
        # self.mongo_conn.pool = self.pool

    def _validate_query(self, query: dict):
        """ Validates a query. Must have `type` and a second identifier at least"""
        if 'type' not in query:
            return False
        if not isinstance(query['type'], str):
            return False
        if len(query) < 2:
            return False
        return True

    def _generate_hash(self, query: dict):
        _hash = ujson.dumps(query, sort_keys=True)
        _hash = base64.b64encode(str.encode(_hash))
        _hash = _hash.decode('utf-8')
        return _hash

    def _check_redis_for_prior(self, _hash: str) -> bool:
        """ Checks to see if any """
        prior_length = self.redis.llen(_hash)
        if prior_length == 0:
            return False
        return True

    def _update_dict(self, query: dict, data: dict):
        query = copy(query)
        timestamp = maya.now()._epoch
        query['timestamp'] = timestamp
        data.update(query)
        return data

    def _update_dict_no_timestamp(self, query: dict, data: dict):
        query = copy(query)
        data = copy(data)
        data.update(query)
        data.pop("timestamp", None)
        return data

    def _omit_timestamp(self, data: dict):
        """ Removes timestamp if it exists. Use it to create a copied version of a dictionary to be saved in the duplicate list """
        _data = copy(data)
        _data.pop("timestamp", None)
        return _data

    def back_to_dict(self, list_of_serialized: list):
        deserialized = []
        if len(list_of_serialized) == 1:
            return orjson.loads(list_of_serialized[0])

        for i in list_of_serialized:
            deserialized.append(orjson.loads(i))
        return deserialized

    def _save(self, query: dict, data: dict):
        """
            Given a type (data entity), data and a epoch for time (utc time only), save the data in both redis and mongo. 
            Does it in a background process. Use with add event.
            We save the information both in mongodb and redis. We assume there's many of each collection. We find a specific collection using the query.
        """
        self.redis_conn.save(query, data)
        self.pool.schedule(self.mongo_conn.save, args=(query, data))

    """
        RESET FUNCTIONS
    """

    def _reset_count(self, query: dict):
        """ Reset the count for the current mongodb query. We do this by adding records in mongo back into redis. """
        all_elements = self.mongo_conn.query_all(query)
        self.pool.schedule(self.redis_conn.reset, args=(query, all_elements))

    def reset(self, query: dict):
        """ Resets all of the variables """
        self.pool.schedule(self._reset_count, args=(query))

    """
        DELETES FUNCTIONS
    """

    def _remove(self, query: dict, details: dict):
        """ Use to both remove items from redis and mongo. Add it when you need it."""

        """ 
            Removes the given query information from the database. 
            It's a heavy computation on redis, as it'll require searching an entire list.
            
        """
        self.pool.schedule(self.mongo_conn.delete_all, args=(query, details))
        self.redis_conn.delete(query, details)
        self.pool.schedule(self.redis_conn.delete_all, args=(query))

    def _remove_first_redis(self, _hash, query: dict):
        # rlock = f"{_hash}:lock"
        # with self.redis.lock(rlock):
        #     push_key = f"{_hash}:list"
        #     self.redis.rpop(push_key)
        pass

    def remove_first(self, query: dict):
        pass
        # _hash = self._generate_hash(query)
        # count = self._get_count(_hash, query)

        # if count == 0:
        #     return

        # self._remove_first_redis(_hash, query)

    """ 
        SAVE FUNCTIONS
    """

    def save(self, query: dict, data: dict):
        self._save(query, data)

    def save_many(self, query: dict, data: List[dict]):
        if self._validate_query(query) == False:
            # Log a warning here instead
            return

        if len(data) == 0:
            return

        for item in data:
            self._save(query, item)

    def bulk_upsert_redis(self, query, data):
        logger.info("Default retcon redis")
        self.pool.schedule(self.redis_conn.update_many, args=(query, data))

    def _bulk_save(self, query, data: list):
        """ Bulk adds a list to redis."""

        self.redis_conn.save_many(query, data)
        self.pool.schedule(self.mongo_conn.save_many, args=(query, data))

    def _save_redis(self, _hash: str, data: dict):

        serialized = orjson.dumps(data)
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            push_key = f"{_hash}:list"
            self.redis.rpush(push_key, serialized)

    def _bulk_save_redis(self, _hash: str, data: list):
        serialized_list = [orjson.dumps(x) for x in data]

        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            push_key = f"{_hash}:list"
            self.redis.rpush(push_key, *serialized_list)

    def _pop_redis_multiple(self, _hash, limit: int):
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            with self.redis.pipeline() as pipe:
                latest_items = []
                try:
                    push_key = f"{_hash}:list"
                    pipe.watch(push_key)
                    latest_items = pipe.lrange(push_key, -limit, -1)
                    pipe.ltrim(push_key, 0, -limit)
                    pipe.execute()

                except Exception as e:
                    pass
                finally:
                    pipe.reset()
                if len(latest_items) > 0:
                    return self.back_to_dict(latest_items)
                return latest_items

    def pop_multiple(self, query, limit: int = 1):
        """ Get multiple items """
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0:
            return []
        return self._pop_redis_multiple(_hash, limit)

    """
        Public Query Functions
    """

    def query_direct(self, query):
        """ Queries from mongodb directly. Used to search extremely large queries. """
        latest_items = list(self.store.query_latest(query))
        return latest_items

    def query_direct_latest(self, query):
        """ Queries from mongodb directly. Used to search extremely large queries. """
        latest_items = list(self.store.query_latest(query))
        if len(latest_items) > 0:
            return latest_items[0]
        return {}

    def get_latest(self, query):
        """ Gets the latest query"""
        # Add a conditional time lock
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count > 0:
            return self.back_to_dict(self.redis.lrange(f"{_hash}:list", -1, -1))
        # Mongo, slowdown
        latest_items = list(self.store.query_latest(query))
        if len(latest_items) > 0:
            return latest_items[0]
        return {}

    def get_latest_many(self, query: dict, limit=1000):

        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0: return []

        latest_redis_items = self.back_to_dict(self.redis.lrange(f"{_hash}:list", -limit, -1))
        # TODO: Get mongo tasks here
        # How will this work now?
        rlen = len(latest_redis_items)
        if rlen == 0:
            query["limit"] = limit
            latest_items = list(self.store.query_latest(query))
            self.reset(query)
            return latest_items

        return latest_redis_items

    """
        SEARCH ONE FUNCTIONS
    """

    def _search_one(self, item: dict, query: dict):
        all_bools = []
        for q in query:
            if q in item:
                if query[q] == query[q]:
                    all_bools.append(True)
                else:
                    all_bools.append(False)
            else:
                all_bools.append(False)
        return any(all_bools)

    def _get_count(self, _hash: str, query: dict):
        # Checks to see if a count already exist in redis, if not, check for a count in mongo.
        _count_hash = f"{_hash}:list"
        count = self.redis.llen(_count_hash)
        if count is not None:
            return count

        records = list(self.store.query(query))
        record_len = len(records)
        return record_len

    def count(self, query):
        """ """
        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        return count



import maya
from loguru import logger
from typing import Dict, List, Any
from jamboree.storage.databases import DatabaseConnection


class MongoDatabaseConnection(DatabaseConnection):
    def __init__(self) -> None:
        super().__init__()

    """ Save commands """

    def save(self, query: dict, data: dict):
        if not self.helpers.validate_query(query):
            # Log a warning here instead
            return
        timestamp = maya.now()._epoch
        query.update(data)
        query['timestamp'] = timestamp
        self.connection.store(query)

    def save_many(self, query: Dict[str, Any], data: List[Dict]):
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        first_item = data[0]
        first_item.update(query)
        updated_list = [self.helpers.update_dict(query, x) for x in data]
        self.connection.bulk_upsert(updated_list, _column_first=first_item.keys(), _in=['timestamp'])

    """
        Update commands
    """

    def update_single(self, query):
        pass

    def update_many(self, query):
        pass

    """
        Delete Commands
    """

    def delete(self, query: dict, details: dict):
        if not self.helpers.validate_query(query):
            return

        query.update(details)
        self.connection.delete(query)

    def delete_many(self, query: dict, details: dict = {}):
        if not self.helpers.validate_query(query):
            return

        query.update(details)
        self.connection.delete_many(query)

    def delete_all(self, query: dict):
        logger.info("Same as `delete_many`")
        self.delete_many(query)

    """ 
        Query commands
    """

    def query_latest(self, query: dict, abs_rel="absolute"):
        if not self.helpers.validate_query(query):
            return {}
        latest_items = self.connection.query_last(query)
        return latest_items

    def query_latest_many(self, query: dict):
        if not self.helpers.validate_query(query):
            return []
        latest_items = list(self.connection.query_latest(query))
        return latest_items

    def query_all(self, query: dict):
        if not self.helpers.validate_query(query):
            return []
        mongo_data = list(self.connection.query(query))
        return mongo_data
    

    def query_latest_by_time(self, query:dict, max_epoch:float, abs_rel:str="absolute", limit:int=10):
        if not self.helpers.validate_query(query):
            return {}
        latest_items = self.connection.query_closest(query)
        return latest_items

    def query_between(self, query:dict, min_epoch:float, max_epoch:float, abs_rel:str="absolute"):
        if not self.helpers.validate_query(query):
            return {}
        
        latest_items = list(self.connection.query_time(query, time_type="window", start=min_epoch))
        if len(latest_items) == 0:
            return []
        return latest_items
    
    def query_before(self, query):
        if not self.helpers.validate_query(query):
            return []
        mongo_data = list(self.connection.query(query))
        return mongo_data
    
    def query_after(self, query):
        if not self.helpers.validate_query(query):
            return []
        mongo_data = list(self.connection.query(query))
        return mongo_data

    """ Swap focused commands"""

    def query_mix_swap(self):
        pass

    def swap(self):
        pass

    """ 
        Pop commands
    """

    def pop(self, query: dict):
        if not self.helpers.validate_query(query):
            return []

        query['limit'] = 1
        item = list(self.connection.query_latest(query))
        if item is not None:
            self.connection.delete(item)
        return item

    def pop_many(self, query: dict, limit: int = 10):
        if not self.helpers.validate_query(query):
            return []

        query['limit'] = limit
        items = list(self.connection.query_latest(query))
        if len(items) == 0:
            return []
        for item in items:
            self.connection.delete(item)
        return items

    def get_latest_many_swap(self):
        pass

    """ Other Functions """

    def reset(self, query: dict):
        pass

    def count(self, query: dict):
        if not self.helpers.validate_query(query):
            return 0
        query.pop('limit', None)
        records = list(self.connection.query(query))
        record_len = len(records)
        return record_len



# import random
# import time
# import uuid
# from pprint import pprint
# from typing import Any, List, Optional

# import maya
# from addict import Dict
# from loguru import logger

# from jamboree import Jamboree
# from jamboree.handlers.complex.engines import FileEngine
# from jamboree.middleware.procedures import (ModelProcedureAbstract,
#                                             ModelProcedureManagement,
#                                             ProcedureAbstract,
#                                             ProcedureManagement)
# from jamboree.utils.support.search import querying

# logger.disable(__name__)

# class ModelEngine(FileEngine):
#     """ """

#     def __init__(self, processor, **kwargs):
#         super().__init__(processor=processor, **kwargs)
#         self.pm = ModelProcedureManagement()
#         self.current_procedure = None

#     def init_specialized(self, **kwargs):
#         super().init_specialized(**kwargs)
#         self.model_type = kwargs.get("model_type", "sklearn")

#     def open_context(self):
#         if not self.file_reset:
#             self.reset()
        

#     def close_context(self):
#         current_model = self.model
#         if current_model.changed:
#             extracted = current_model.extract()
#             self.save_file(extracted)
        
#         # probably do some sort of metrics stuff :) 


#     def enterable(self):
#         """ Return the object we want to enter into """
#         return self.model

#     def custom_post_load(self, item):
#         proc = self.procedure
#         proc.dictionary = item
#         proc.verify()
#         self.current_procedure = proc
    

#     @property
#     def procedure(self) -> 'ModelProcedureAbstract':
#         if not self.current_procedure:
#             self.current_procedure = self.pm.access(self.model_type)
#             logger.success(f"Successfully accessed a procedure: {self.current_procedure}")
#         return self.current_procedure
    
#     @procedure.setter
#     def procedure(self, _procedure:'ModelProcedureAbstract'):
#         self.current_procedure = _procedure

#     @property
#     def model(self):
#         if self.current_procedure:
#             self.procedure.verify()
#             return self.procedure
#         raise AttributeError("You haven't added a procedure yet")
    
#     def file_from_dict(self, item:Dict):
#         reloaded = ModelEngine(
#             processor=self.processor,
#             name=item.name,
#             category=item.category,
#             subcategories=item.subcategories,
#             submetatype=item.submetatype,
#             abbreviation=item.abbreviation,
#             model_type=item.submetatype
#         )
#         return reloaded


#     def reset(self):
#         super().reset()


# def file_engine_main():
#     """ 
#         Creating a generic usage of the file engine instead of only model storage. 

#         To test, we're going to entirely duplicate the prior test. 
#         Only we're going to use generic functions and variables. In essensce, rebuild the `ModelEngine` starting with the file handler

#     """

#     from jamboree.middleware.procedures.models import CustomSklearnGaussianProcedure
#     file_name = uuid.uuid4().hex
#     logger.info("Starting file engine experiment")
#     logger.info(f"The file name is: {file_name}")
#     jamboree_processor = Jamboree()
#     with logger.catch(message="There should be no reason for this to fail"):
#         # Initialize a file engine
#         model_engine = ModelEngine(
#             processor=jamboree_processor,
#             name=file_name,
#             category="machine",
#             subcategories={"ml_type": "gaussian"},
#             abbreviation="GAUSS",
#             submetatype="sklearn",
#             blobfile=CustomSklearnGaussianProcedure(),
#         )
#         model_engine.reset()

#         reloaded = model_engine.first(name=file_name)
#         while True:
#             with reloaded as model:
#                 logger.debug(model)


# if __name__ == "__main__":
#     logger.enable(__name__)
#     file_engine_main()



import maya
import orjson
from copy import copy

from redis.client import Pipeline, Redis
from typing import Dict, List, Optional, Any, AnyStr, Union
from pprint import pprint
from jamboree.storage.databases import DatabaseConnection

# from redis.exceptions import WatchError
from jamboree.utils.context import watch_loop, watch_loop_callback

# NOTE: Can add pipelining to redis storage to make fewer calls.
from loguru import logger
import redis

# def default()


class RedisDatabaseZSetsConnection(DatabaseConnection):
    """ 
        Redis Connection Notes
        ---
        We're going to end up using zadds and zscores. 
        The indexable score makes it a good target for time series. 
        The fact that it's a set makes it so there can be no duplicates added.

        In essense. Perfect for time series. 
    """

    def __init__(self) -> None:
        super().__init__()

    """
        # Individual Access Methods
        ---
        Use commands here to have single variable in place instead of creating an event log. 
        Very useful for stateful applications such as timeindexing logic.

        ## Core Functions:

        * kill - delete a given hash/key
        * add  - set a given hash/key to a variable. Usually that should be a dictionary representing a complex value.
        * get  - get a variable by hash/key
    """

    def _kill(self, _hash: str):
        """ Deletes a key within a lock"""
        rlock = f"{_hash}:lock"
        sub_key = f"{_hash}:single"
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.delete(sub_key)
            pipe.execute()

    def _add(self, _hash: str, data: dict, is_serialized=True):
        """Set a single variable to the redis database using the redis set command. 

        Arguments:
            _hash {str} -- The hash identifier
            data {dict} -- The data we're going to add.

        Keyword Arguments:
            is_serialized {bool} -- Determine if we're to serialize the underlying variable. (default: {False})
        """
        rlock = f"{_hash}:lock"
        sub_key = f"{_hash}:single"
        serialized = None
        if is_serialized:
            # Don't serialize the json
            serialized = orjson.dumps(data)
        else:
            serialized = data
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.set(sub_key, serialized)
            pipe.execute()

    def _get(self, _hash: str):
        sub_key = f"{_hash}:single"
        value = None
        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(sub_key)
                    value = pipe.get(sub_key)
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue

        return value

    def get(self, query: dict, is_serialized=True):
        if not self.helpers.validate_query(query):
            return {}

        _hash = self.helpers.generate_hash(query)
        value = self._get(_hash)
        if value is not None:
            if is_serialized:
                return orjson.loads(value)
            return value
        return {}

    def add(self, query: dict, data: dict, is_serialized=True):
        """ Sets a single value. It's a wrapper around"""
        if not self.helpers.validate_query(query):
            if is_serialized:
                if len(data) == 0:
                    return

        _hash = self.helpers.generate_hash(query)
        self._add(_hash, data, is_serialized=is_serialized)

    def kill(self, query: dict):
        """ Deletes a single variable. Bypasses the stack"""
        if not self.helpers.validate_query(query):
            return

        _hash = self.helpers.generate_hash(query)
        self._kill(_hash)

    """ 
        # Save Commands
        ---
        * `save` - ...
        * `save_many` - ...
    """

    @logger.catch
    def _save(self, _hash: str, data: dict, timing: dict):
        """ Appends an event to the stack. """
        serialized = orjson.dumps(data)
        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                relative_data = {serialized: timing["time"]}
                absolute_data = {serialized: timing["timestamp"]}
                pipe.zadd(relative_time_key, relative_data)
                pipe.zadd(absolute_time_key, absolute_data)
            pipe.execute()

    def save(self, query: dict, data: dict, _time=None, _timestamp=None):
        """ Save a single record. """
        if not self.helpers.validate_query(query):
            return
        _hash = self.helpers.generate_hash(query)
        query.update(data)
        data, timing = self.helpers.separate_time_data(query, _time, _timestamp)
        # print(timing)
        self._save(_hash, data, timing)

    def _save_many(self, _hash: str, relative_data: Dict[str, float] = {}):
        # serialized_list = [orjson.dumps(x) for x in data]

        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        absolute_data = self.helpers.get_current_abs_time(relative_data)
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.zadd(relative_time_key, relative_data)
                pipe.zadd(absolute_time_key, absolute_data)
            pipe.execute()

    def save_many(self, query, data: Dict[str, float] = {}, abs_rel="absolute"):
        """ 
            # Save Many

            Save multiple records at once. We assume the data was already processed. 
        """
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        _hash = self.helpers.generate_hash(query)
        self._save_many(_hash, data)

    """ 
        # Delete Commands
        ---
        * `delete` - Get all of the records related to a given key.
        * `delete_latest` - Get the `n` latest records according to our query parameters.
        * `delete_between` - Query between two epoch times.
        * `delete_before` - Get everything before an epoch time.
        * `delete_after` - Get everything after epoch time.

    """

    def delete_first(self, query, details):
        pass

    def _delete(self, _hash: str, details: dict):
        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        deletion_key = orjson.dumps(details)
        with self.connection.lock(rlock):
            self.connection.zrem(relative_time_key, deletion_key)
            self.connection.zrem(absolute_time_key, deletion_key)

    def delete(self, query: dict, details: dict):
        if not self.helpers.validate_query(query):
            return
        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0:
            return
        self._delete(_hash, details)

    def _delete_many(self, _hash: str, data: List[Dict]):
        pass

    def delete_many(self, query, data: List[Dict]):
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        updated_list = [self.connection.update_dict(query, x) for x in data]
        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0:
            return
        self._delete_many(_hash, updated_list)

    def _delete_all(self, _hash: str):
        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"

        with self.connection.lock(rlock):
            keys = self.connection.zrange(relative_time_key, 0, -1, withscores=True)

            values = [key[0] for key in keys]
            if len(values) == 0:
                return

            self.connection.zrem(relative_time_key, *values)
            self.connection.zrem(absolute_time_key, *values)

    def delete_all(self, query: dict):
        if not self.helpers.validate_query(query):
            return

        _hash = self.helpers.generate_hash(query)
        self._delete_all(_hash)

    """
        Query commands
        ---
        All of the query information we need to operate on.


        The query functions we work on:

        1. `query` - Get all of the records related to a given key.
        2. `query_latest` - Get the `n` latest records according to our query parameters.
        3. `query_between` - Query between two epoch times.
        4. `query_before` - Get everything before an epoch time.
        5. `query_after` - Get everything after epoch time.    
    """

    def query_all(self, query: dict):
        """ Same as query_all """
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        with self.connection.pipeline() as pipe:
            akeys = None
            rkeys = None
            while True:
                try:
                    pipe.watch(relative_time_key)
                    pipe.watch(absolute_time_key)
                    count = self.count(_hash, pipe=pipe)
                    if count == 0:
                        return []

                    rkeys = pipe.zrange(relative_time_key, 0, -1, withscores=True)
                    akeys = pipe.zrange(absolute_time_key, 0, -1, withscores=True)
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue
        combined = self.helpers.combine_results(akeys, rkeys)
        return combined

    def query_latest(self, _query, abs_rel="absolute", limit: int = 10):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return {}
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)

        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(_current_key)
                    count = self.count(_hash, pipe=pipe)
                    if count == 0:
                        return {}

                    keys = pipe.zrange(_current_key, -1, -1, withscores=True)
                    if len(keys) == 0:
                        return {}
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue
        combined = self.helpers.combined_abs_rel(keys, abs_rel=abs_rel)
        return combined[-1]

    def query_latest_many(self, _query, abs_rel="absolute", limit: int = 10):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return {}
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(_current_key)
                    count = self.count(_hash, pipe=pipe)
                    if count == 0:
                        return {}

                    keys = pipe.zrange(_current_key, -limit, -1, withscores=True)
                    if len(keys) == 0:
                        return {}
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue

        combined = self.helpers.combined_abs_rel(keys, abs_rel=abs_rel)
        return combined

    def query_between(
        self,
        _query: dict,
        min_epoch: float,
        max_epoch: float,
        abs_rel: str = "absolute",
    ):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)

        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(_current_key)
                    count = self.count(_hash, pipe=pipe)
                    if count == 0:
                        return []

                    keys = pipe.zrangebyscore(
                        _current_key, min=min_epoch, max=max_epoch, withscores=True
                    )
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue
        combined = self.helpers.combined_abs_rel(keys, abs_rel=abs_rel)
        return combined


    def query_all_between(
        self,
        _query: dict,
        abs_rel: str = "absolute",
    ):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)

        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(_current_key)
                    count = self.count(_hash, pipe=pipe)
                    min_epoch = self.min_score(_hash, pipe=pipe)
                    max_epoch = self.max_score(_hash, pipe=pipe)
                    if count == 0:
                        return []

                    keys = pipe.zrangebyscore(
                        _current_key, min=min_epoch, max=max_epoch, withscores=True
                    )
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue
        combined = self.helpers.combined_abs_rel(keys, abs_rel=abs_rel)
        return combined

    def query_latest_by_time(
        self, _query, max_epoch, abs_rel="absolute", limit: int = 10
    ):
        """ Get the closest item to a given timestamp. Simply pass in an epoch then watch things fly"""
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(_current_key)
                    count = self.count(_hash, pipe=pipe)
                    if count == 0:
                        return []
                    keys = pipe.zrangebyscore(
                        _current_key,
                        min=max_epoch,
                        max="+inf",
                        start=0,
                        num=1,
                        withscores=True,
                    )
                    combined = self.helpers.combined_abs_rel(keys, abs_rel)
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue
        if len(combined) == 0:
            return {}
        return combined[0]

    def query_before(self, _query, max_epoch, abs_rel="absolute"):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(_current_key)
                    count = self.count(_hash, pipe=pipe)
                    if count == 0:
                        return []
                    keys = pipe.zrangebyscore(
                        _current_key, min="-inf", max=max_epoch, withscores=True
                    )
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue
        combined = self.helpers.combined_abs_rel(keys, abs_rel)
        return combined

    def query_after(self, _query, min_epoch, abs_rel="absolute"):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        with self.connection.pipeline() as pipe:
            while True:
                try:
                    count = self.count(_hash, pipe=pipe)
                    if count == 0:
                        return []
                    keys = pipe.zrangebyscore(
                        _current_key, min=min_epoch, max="+inf", withscores=True
                    )
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue

        combined = self.helpers.combined_abs_rel(keys, abs_rel)
        return combined

    """
        Other Functions
    """

    def _reset_count(self, query: dict, mongo_data: List[Dict]):
        """ Reset the count for the current mongodb query. We do this by adding records in mongo back into redis. """
        if len(mongo_data) == 0:
            return

        _hash = self.helpers.generate_hash(query)
        phindex = self.connection.incr("placeholder:index")
        delindex = self.connection.incr("deletion:index")
        _hash_key = f"{_hash}:list"
        _hash_placeholder = f"{_hash}:{phindex}"
        _hash_del = f"{_hash}:{delindex}"

        serialized_mongo = [orjson.dumps(mon) for mon in mongo_data]
        # rlock = f"{_hash}:lock"
        with self.connection.pipeline() as pipe:
            while True:
                try:
                    pipe.watch(_hash)
                    pipe.watch(_hash_placeholder)
                    pipe.watch(_hash_key)
                    pipe.connection.rpush(_hash_placeholder, *serialized_mongo)
                    pipe.connection.rename(_hash_key, _hash_del)
                    pipe.connection.rename(_hash_placeholder, _hash_key)
                    pipe.execute()
                    break
                except redis.exceptions.WatchError:
                    continue

        self._delete_all(_hash_del)

    def reset(self, query, mongo_data: List[Dict]):
        if not self.helpers.validate_query(query):
            return
        self.pool.schedule(self._reset_count, args=(query, mongo_data))

    def count(self, _hash: str, pipe=None) -> int:
        # ZCARD to get the length of the zset
        _count_hash = f"{_hash}:alist"
        if pipe is not None:
            pipe.watch(_count_hash)
            count = pipe.zcard(_count_hash)
        else:
            count = self.connection.zcard(_count_hash)
        return int(count)

    def general_lock(self, query: dict):
        """ 
            Use this lock to wrap around step functions. 
            It would prevent race conditions on the same object (portfolio manipulation)
        """
        _hash = self.helpers.generate_hash(query)
        general_key = f"{_hash}:lock:generalized"
        return self.connection.lock(general_key)

    def min_score(self, _hash: str, pipe: Optional[Pipeline] = None):
        """max_score Get Max Score

            Get the max score given a key. We get the maximum score and cache it. 
            If we change any information we can swap the cache out dynamically to increase access speed.
        """
        _count_hash = f"{_hash}:rlist"
        if pipe is not None:
            pipe.watch(_count_hash)
            count = pipe.zrangebyscore(
                start=0, num=1,
                name=_count_hash,
                min="-inf", max="+inf", 
                withscores=True
            )

        else:
            count = self.connection.zrangebyscore(
                name=_count_hash,
                start=0, num=1,
                min="-inf", max="+inf", 
                withscores=True
            )
        if count is None:
            return float(0.0)

        if len(count) > 0:
            return float(count[0][1])
        return float(0.0)


    # def max_score(self, _hash: str, pipe: Optional[Pipeline] = None):
    #     """max_score Get Max Score

    #         Get the max score given a key. We get the maximum score and cache it. 
    #         If we change any information we can swap the cache out dynamically to increase access speed.
    #     """
    #     _count_hash = f"{_hash}:rlist"
    #     if pipe is not None:
    #         pipe.watch(_count_hash)
    #         count = pipe.zrangebyscore(
    #             start=0, num=1,
    #             name=_count_hash,
    #             min="-inf", max="+inf", 
    #             withscores=True
    #         )

    #     else:
    #         count = self.connection.zrangebyscore(
    #             name=_count_hash,
    #             start=0, num=1,
    #             min="-inf", max="+inf", 
    #             withscores=True
    #         )
    #     if count is None:
    #         return float(0.0)

    #     if len(count) > 0:
    #         return float(count[0][1])
    #     return float(0.0)

    def max_score(self, _hash: str, pipe: Optional[Pipeline] = None):
        """min_score Get Minimum Score

            Get the min score given a key. We get the maximum score and cache it. 
            If we change any information we can swap the cache out dynamically to increase access speed.
        """
        _count_hash = f"{_hash}:alist"
        first_dict: Optional[Dict[AnyStr, Any]] = None

        if pipe is not None:
            pipe.watch(_count_hash)
            first_dict = pipe.zrangebyscore(
                name=_count_hash,
                start=0, num=1,
                min="+inf", max="-inf", 
                withscores=True
            )
        else:
            first_dict = self.connection.zrangebyscore(
                name=_count_hash,
                start=0, num=1,
                min="+inf", max="-inf",
                withscores=True
            )

        if first_dict is None:
            return float(0.0)

        if len(first_dict) > 0:
            return float(first_dict[0][1])
        return float(0.0)



def dict_validation(obj:dict) -> bool:
    obj_keys = list(obj.keys())
    for x in ['subcategories', 'entity', 'submetatype', 'name', 'metatype', 'category', "abbreviation"]:
        if x not in obj_keys:
            return False
    return True

def default(obj):
    pass


from jamboree.handlers.default import DataHandler

class EconomicData(DataHandler):
    """
        # Economic Data
        
        A way to browse economic data. Is an extension of DataHandler and includes basic searches.
    """
    pass


import lz4.frame
import dill

""" 
    # COMPRESSED SERIALIZATION LIBRARY

    Simply compress and serialize
"""

def serialize(obj):
    """ Should take a complex object and pickle it"""
    pickled = dill.dumps(obj, byref=False)
    compressed = lz4.frame.compress(pickled)
    return compressed

def deserialize(obj):
    """ Should take a serialized object and pickle"""
    decompressed = lz4.frame.decompress(obj)
    unpickled = dill.loads(decompressed)
    return unpickled


import uuid
from loguru import logger

import maya
import pandas as pd

from jamboree import Jamboree
from jamboree import JamboreeNew

from jamboree.handlers.complex.meta import MetaHandler
# from jamboree.handlers.complex.metric import MetricHandler

from jamboree.handlers.default.time import TimeHandler
from jamboree.handlers.default.db import DBHandler
from jamboree.handlers.default import BlobStorageHandler

class BacktestBlobHandler(BlobStorageHandler):
    def __init__(self):
        super().__init__()
        
        
        # Other objects to consider
        self._time:TimeHandler = TimeHandler()
        self._meta:MetaHandler = MetaHandler()
        self._episode = uuid.uuid4().hex
        
        
        self._is_live = False
        self.is_event = False # use to make sure there's absolutely no duplicate data
    
    # Create a context for this to remove complete dependency from the model handler.
    
    @property
    def episode(self) -> str:
        return self._episode
    
    @episode.setter
    def episode(self, _episode:str):
        self._episode = _episode
    
    @property
    def live(self) -> bool:
        return self._is_live
    
    @live.setter
    def live(self, _live:bool):
        self._is_live = _live

    @property
    def time(self) -> 'TimeHandler':
        # self._time.event = self.event
        self._time.processor = self.processor
        self._time['episode'] = self.episode
        self._time['live'] = self.live
        return self._time

    @time.setter
    def time(self, _time:'TimeHandler'):
        self._time = _time


    def reset(self):
        """ Reset the data we're querying for. """
        self.time.reset()
    
    
    
    def __str__(self) -> str:
        name = self["name"]
        category = self["category"]
        subcategories = self["subcategories"]
        jscat = self.main_helper.generate_hash(subcategories)
        return f"{name}:{category}:{jscat}"


from .base.core import Jamboree
from .base.processors.main import Jamboree as JamboreeNew
from .handlers.base import BaseHandler
from .handlers.default.db import DBHandler
from .handlers.default.data import DataHandler
from .handlers.default.time import TimeHandler


from abc import ABC, ABCMeta
import copy

from typing import Dict, Any, List
from loguru import logger
from .processor import EventProcessor


class BaseHandler(object, metaclass=ABCMeta):
    """ 
        A way to handle reads and writes consistently without having to write every single variable:
    """

    def __init__(self):
        pass

    def check(self):
        raise NotImplementedError

    def save(self, data: dict):
        raise NotImplementedError

    def _bulk_save(self, query: dict, data: list):
        raise NotImplementedError

    def _get_many(self):
        raise NotImplementedError

    def last(self):
        raise NotImplementedError

    def many(self, limit: int = 100):
        raise NotImplementedError

    def save_many(self, query: dict, data: list):
        raise NotImplementedError

    def pop_multiple(self, query, _limit: int = 1):
        raise NotImplementedError

    def swap(self, query, alt: dict = {}):
        """ Swap betwen the first and last item """
        raise NotImplementedError

    def query_mix(self, query: dict, alt: dict = {}):
        raise NotImplementedError


class DBHandler(BaseHandler):
    """ 
        A way to handle reads and writes consistently without having to write every single variable:
    """

    def __init__(self):
        # print("DBHandler")
        self._entity = ""
        self._required = {}
        self._query = {}
        self.data = {}
        self.event_proc = None

    def __setitem__(self, key, value):
        if bool(self.required):
            if key in self.required:
                self._query[key] = value
                return self._query
        self.data[key] = value
        return self.data

    def __getitem__(self, key):
        if key in self._query.keys():
            return self._query.get(key, None)
        else:
            if key in self.data.keys():
                return self.data.get(key, None)
        return None

    @property
    def event(self):
        return self.event_proc

    @event.setter
    def event(self, _event: EventProcessor):
        # Use to process event
        self.event_proc = _event

    @property
    def entity(self):
        return self._entity

    @entity.setter
    def entity(self, _entity: str):
        self._entity = str(_entity)

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, _required: Dict[str, Any]):
        # check to make sure it's not empty
        self._required = _required

    @property
    def query(self):
        return self._query

    @required.setter
    def query(self, _query: Dict[str, Any]):
        if len(_query.keys()) > 0:
            self._query = _query

    def check(self):
        if self.event_proc is None:
            raise AttributeError("Event processor isn't available.")

        if (not bool(self._entity)) or (not bool(self._required)) or (not bool(self._query)):
            raise AttributeError(f"One of the key variables is missing.")

        for req in self._required.keys():
            _type = self._required[req]
            if req not in self._query:
                raise AttributeError(f"{req} is not in the requirements")
            if not isinstance(self._query[req], _type):
                raise AttributeError(f"{req} is not a {_type}")
        return True

    def save(self, data: dict, alt={}):
        self.check()

        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        self.event_proc.save(query, data)

    def save_many(self, data: list, alt={}):
        self.check()

        query = copy.copy(self._query)
        query['type'] = self.entity
        # logger.info(query)
        query.update(alt)
        query.update(self.data)
        self.event_proc._bulk_save(query, data)

    def _get_many(self, limit: int, alt={}):
        """ Aims to get many variables """
        self.check()
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        latest_many = self.event_proc.get_latest_many(query, limit=limit)
        return latest_many

    def _get_latest(self, alt={}):
        self.check()
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        latest = self.event_proc.get_latest(query)
        return latest

    def last(self, alt={}):
        alt.update(self.data)
        return self._get_latest(alt)

    def many(self, limit=1000, alt={}):
        alt.update(self.data)
        return self._get_many(limit, alt=alt)

    def pop(self, alt={}):
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        self.event_proc.remove_first(query)

    def pop_many(self, _limit, alt={}):
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        return self.event_proc.pop_multiple(query, _limit)

    def count(self, alt={}):
        """ Aims to get many variables """
        self.check()
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        return self.event_proc.count(query)

    def swap_many(self, limit: int = 10, alt={}):
        """ Move items from the main list to a swapped list. """
        self.check()
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        return self.event_proc.multi_swap(query, limit)

    def query_mix(self, limit: int = 10, alt: dict = {}):
        self.check()
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        return self.event_proc.query_mix(query, limit)

    def query_many_swap(self, limit: int = 10, alt: dict = {}):
        self.check()
        query = copy.copy(self._query)
        query['type'] = self.entity
        query.update(alt)
        query.update(self.data)
        return self.event_proc.get_latest_many_swap(query, limit)



from .database import DatabaseConnection
from .jmongo import MongoDatabaseConnection
from .jredis import RedisDatabaseConnection
from .jredis_zset import RedisDatabaseZSetsConnection as ZRedisDatabaseConnection 


import unittest


"""
    The things tested in this document:
    ---
    1. Storing a lot of information at a single level
        * This is a save_many command on a dataframe
    2. Querying information and resampling it for processing it.
    3. Initializing a certain set of data to process at different layers.
        * This entails getting small data and resampling it prior to querying
        * Query resampled data
"""





from jamboree.handlers.abstracted.search import ParameterizedSearch
from jamboree.handlers.default.search import BaseSearchHandler


class MetadataSearchHandler(ParameterizedSearch):
    """
        # 10 metatypes

        Metatypes are
        
        1. Strategy
        2. Data
        3. Model
        4. Meta

    """
    def __init__(self):
        super().__init__()
        self.entity = "metadata"
        self.dreq = {
            "name": str,
            # The type of metadata we're positioning
            # strategy, data, model, metainfo (metadata about metadata) are the clear items in mind
            "metatype": str,
            # another identifiable metatype to narrow down results
            # pricing(data), economic, weather (data), social(data), political (data), features (data)
            # batch (models), online (models), micro (strategies), macro (strategies), supporting_group (complex)
            "submetatype": str,
            "category": str,
            "subcategories": dict,
            "description": str,
            "info": dict,
            # The location about the information involved
            "location": "GEO",
            "abbreviation": str
        }
        self.must_have = ['name', 'metatype', 'category', 'submetatype', 'abbreviation']
    
    



from contextlib import ContextDecorator, contextmanager

import maya
import redis
from loguru import logger
from redis.exceptions import WatchError


class timecontext(ContextDecorator):
    def __enter__(self):
        self.start = maya.now()._epoch
        return self

    def __exit__(self, *exc):
        self.end = maya.now()._epoch
        delta = self.end - self.start
        logger.success(f"It took {delta}s")
        logger.success(f"It took {(delta*1000)}ms")
        return False

@contextmanager
def watch_loop():
    while True:
        try:
            yield
            break
        except WatchError:
            continue

def watch_loop_callback(callback):
    while True:
        try:
            callback()
            break
        except WatchError:
            continue

class example_space(ContextDecorator):
    def __init__(self, name) -> None:
        self.name = name
        self.is_pass = True
        self.start = maya.now()._epoch
    
    def __enter__(self):
        return self
    
    def failed(self):
        self.is_pass = False

    def __exit__(self, type, value, traceback):
        self.end = maya.now()._epoch
        delta = self.end - self.start
        if value is not None or self.is_pass == False:
            logger.error("----------------------------------------- Example didn't pass --------------------------------------------")
        else:
            logger.success("------------------------------------------ Example did pass ----------------------------------------------")
        logger.info(f"It took {delta}ms")
        return False

if __name__ == "__main__":
    with example_space("Printing") as example:
        print("Don't want to kill my vibe")
        # example.failed()



from abc import ABC
from typing import List, Optional
from jamboree.base.processors.abstracts import EventProcessor, FileProcessor, SearchProcessor


class Processor(ABC):
    """ Use to allow for multiple items"""
    def __init__(self):
        self._event:Optional[EventProcessor] = None
        self._storage:Optional[FileProcessor] = None
        self._search:Optional[SearchProcessor] = None

    @property
    def event(self) -> EventProcessor:
        if not isinstance(self._event, EventProcessor):
            raise AttributeError("EventProcessor not added yet ... ")
        return self._event
    
    @event.setter
    def event(self, _event:EventProcessor) -> EventProcessor:
        self._event = _event
    

    @property
    def storage(self) -> FileProcessor:
        if not isinstance(self._storage, FileProcessor):
            raise AttributeError("FileProcessor not added yet ... ")
        return self._storage
    
    @storage.setter
    def storage(self, _storage:FileProcessor):
        self._storage = _storage

    @property
    def search(self) -> SearchProcessor:
        if not isinstance(self._search, SearchProcessor):
            raise AttributeError("SearchProcessor not added yet ... ")
        return self._search
    
    def search(self, _search:SearchProcessor):
        self._search = self.search


from copy import copy
import orjson
import ujson
from typing import List
from redis import Redis
from funtime import Store
from pebble.pool import ThreadPool
import base64
from multiprocessing import cpu_count
from jamboree.storage.databases import MongoDatabaseConnection, ZRedisDatabaseConnection
from jamboree.utils.helper import Helpers
from jamboree.base.processors.abstracts import EventProcessor
from jamboree.base.processors.abstracts import LegacyProcessor




class Jamboree(EventProcessor):
    """Adds and retrieves events at extremely fast speeds. Use to handle portfolio and trade information quickly."""

    def __init__(self, mongodb_host="localhost", redis_host="localhost", redis_port=6379):
        self.redis = Redis(redis_host, port=redis_port)
        self.store = Store(mongodb_host).create_lib('events').get_store()['events']
        self.pool = ThreadPool(max_workers=cpu_count() * 6)
        self.mongo_conn = MongoDatabaseConnection()
        self.redis_conn = ZRedisDatabaseConnection()
        self.mongo_conn.connection = self.store
        self.redis_conn.connection = self.redis
        self.dominant_database = ""
        self.helpers = Helpers()
        # self.redis_conn.pool = self.pool
        # self.mongo_conn.pool = self.pool

    def _validate_query(self, query: dict):
        """ Validates a query. Must have `type` and a second identifier at least"""
        if 'type' not in query:
            return False
        if not isinstance(query['type'], str):
            return False
        if len(query) < 2:
            return False
        return True

    def _generate_hash(self, query: dict):
        _hash = ujson.dumps(query, sort_keys=True)
        _hash = base64.b64encode(str.encode(_hash))
        _hash = _hash.decode('utf-8')
        return _hash

    def _check_redis_for_prior(self, _hash: str) -> bool:
        """ Checks to see if any """
        prior_length = self.redis.llen(_hash)
        if prior_length == 0:
            return False
        return True

    

    def _omit_timestamp(self, data: dict):
        """ Removes timestamp if it exists. Use it to create a copied version of a dictionary to be saved in the duplicate list """
        _data = copy(data)
        _data.pop("timestamp", None)
        return _data

    def back_to_dict(self, list_of_serialized: list):
        deserialized = []
        if len(list_of_serialized) == 1:
            return orjson.loads(list_of_serialized[0])

        for i in list_of_serialized:
            deserialized.append(orjson.loads(i))
        return deserialized

    def _save(self, query: dict, data: dict, abs_rel="absolute"):
        """
            # TODO: Readd mongo
            Given a type (data entity), data and a epoch for time (utc time only), save the data in both redis and mongo. 
            Does it in a background process. Use with add event.
            We save the information both in mongodb and redis. We assume there's many of each collection. We find a specific collection using the query.
        """
        self.redis_conn.save(query, data)
        # self.pool.schedule(self.mongo_conn.save, args=(query, data))

    """
        RESET FUNCTIONS
    """

    def _reset_count(self, query: dict):
        """ Reset the count for the current mongodb query. We do this by adding records in mongo back into redis. """
        # all_elements = self.mongo_conn.query_all(query)
        # self.pool.schedule(self.redis_conn.reset, args=(query, all_elements))

    def reset(self, query: dict):
        """ Resets all of the variables """
        # self.pool.schedule(self._reset_count, args=(query))

    """
        DELETES FUNCTIONS
    """

    def _remove(self, query: dict, details: dict):
        """ Use to both remove items from redis and mongo. Add it when you need it."""

        """ 
            Removes the given query information from the database. 
            It's a heavy computation on redis, as it'll require searching an entire list.
            
        """
        # self.pool.schedule(self.mongo_conn.delete_all, args=(query, details))
        self.redis_conn.delete(query, details)
        self.pool.schedule(self.redis_conn.delete_all, args=(query))

    def _remove_first_redis(self, _hash, query: dict):
        pass

    def remove_first(self, query: dict):
        pass
        # _hash = self._generate_hash(query)
        # count = self._get_count(_hash, query)

        # if count == 0:
        #     return

        # self._remove_first_redis(_hash, query)

    """ 
        SAVE FUNCTIONS
    """

    def save(self, query: dict, data: dict, abs_rel="absolute"):
        self._save(query, data)

    def save_many(self, query: dict, data: List[dict]):
        if self._validate_query(query) == False:
            # Log a warning here instead
            return

        if len(data) == 0: return
        data_list = [self.helpers.update_dict(query, item) for item in data]
        self._bulk_save(query, data_list)


    def _bulk_save(self, query, data: list):
        """ Bulk adds a list to redis."""
        events = self.helpers.convert_to_storable_relative(data)
        self.redis_conn.save_many(query, events)
        # self.pool.schedule(self.mongo_conn.save_many, args=(query, data))


    def _pop_redis_multiple(self, _hash, limit: int):
        return []

    def pop_multiple(self, query, limit: int = 1):
        """ Get multiple items """
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0:
            return []
        return self._pop_redis_multiple(_hash, limit)

    """
        Public Query Functions
    """

    def query_direct(self, query):
        """ Queries from mongodb directly. Used to search extremely large queries. """
        latest_items = list(self.store.query_latest(query))
        return latest_items

    def query_direct_latest(self, query):
        """ Queries from mongodb directly. Used to search extremely large queries. """
        return self.mongo_conn.query_latest(query)

    def get_latest(self, query, abs_rel="absolute"):
        """ Gets the latest query"""
        # Add a conditional time lock
        _hash = self._generate_hash(query)
        count, database = self._get_count(_hash, query)
        if count > 0:
            return self.redis_conn.query_latest(query, abs_rel)
        return {}
        # # Mongo, slowdown
        # return self.mongo_conn.query_latest(query)

    def get_latest_many(self, query: dict, limit=1000, abs_rel="absolute"):

        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0: return []



        latest_redis_items = self.redis_conn.query_latest_many(query, abs_rel=abs_rel, limit=limit)
        return latest_redis_items


    """ New Queries """


    def get_between(self, query:dict, min_epoch:float, max_epoch:float, abs_rel:str="absolute"):
        items = self.redis_conn.query_between(query, min_epoch, max_epoch, abs_rel)
        return items
    

    def get_latest_by(self, query:dict, max_epoch, abs_rel="absolute", limit:int=10):
        item = self.redis_conn.query_latest_by_time(query, max_epoch, abs_rel)
        return item

    """
        SEARCH ONE FUNCTIONS
    """

    def _search_one(self, item: dict, query: dict):
        all_bools = []
        for q in query:
            if q in item:
                if query[q] == query[q]:
                    all_bools.append(True)
                else:
                    all_bools.append(False)
            else:
                all_bools.append(False)
        return any(all_bools)

    def _get_count(self, _hash: str, query: dict):
        # Checks to see if a count already exist in redis, if not, check for a count in mongo.
        count = self.redis_conn.count(_hash)
        if count is not None:
            return count, "redis"

        records = list(self.store.query(query))
        record_len = len(records)
        return record_len, "mongo"

    def count(self, query):
        """ """
        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count, database = self._get_count(_hash, query)
        self.dominant_database = database
        return count

    def single_get(self, query:dict):
        if self._validate_query(query) == False: return {}
        item = self.redis_conn.get(query)
        return item 

    def single_set(self, query:dict, data:dict):
        if self._validate_query(query) == False: return
        self.redis_conn.add(query, data)

    def single_delete(self, query:dict):
        if self._validate_query(query) == False: return
        self.redis_conn.kill(query)


import base64
import ujson
import orjson
from cytoolz import keyfilter


def consistent_hash(query: dict) -> str:
    _hash = ujson.dumps(query, sort_keys=True)
    _hash = base64.b64encode(str.encode(_hash))
    _hash = _hash.decode('utf-8')
    return _hash

def consistent_unhash(_hash:str) -> str:
    """ Take a consistent hash (sorted) and turn it back into a dictionary"""
    decoded_hash = base64.b64decode(_hash).decode('utf-8')
    _hash_dict = ujson.loads(decoded_hash)
    return _hash_dict


def omit(blacklist, d):
    return keyfilter(lambda k: k not in blacklist, d)





from redis import Redis
from jamboree.base.processors.abstracts import Processor
from jamboree.base.processors.event import JamboreeEvents
from jamboree.base.processors.files import JamboreeFileProcessor

class Jamboree(Processor):
    def __init__(self, **kwargs) -> None:
        super().__init__()

        redis_host = kwargs.get("REDIS_HOST", "localhost")
        redis_port = int(kwargs.get("REDIS_PORT", "6379"))
        mongo_host = kwargs.get("MONGO_HOST", "localhost")
        rconn = Redis(host=redis_host, port=redis_port)
        # redis.Redis(redis_host, port=redis_port)

        self.event = JamboreeEvents(
            mongodb_host=mongo_host,
            redis_host=redis_host,
            redis_port=redis_port
        )

        # Set the files management here
        self.storage = JamboreeFileProcessor()
        self.storage.rconn = rconn
        self.event.rconn = rconn
        self.event.initialize()
        self.storage.initialize()
        self.rconn = rconn


import uuid
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import maya
import pprint
from loguru import logger
from typing import List
from jamboree.handlers.default import DataHandler
from jamboree import Jamboree
from jamboree.utils.support.search import querying

class PriceData(DataHandler):
    """
        # Price Data
        
        A way to browse and interact with price data. Is an extension of DataHandler and includes basic searches.

    """
    def __init__(self):
        super().__init__()
        self['category'] = "markets"
        self['submetatype'] = "price"
        self.sc = "subcategories" # storing the placeholder key to prevent misspelling
        # self.cat = "category" # storing variable placeholder key to prevent misspelling

    @property
    def markets(self) -> List[str]:
        return [
            'crypto', 'stock', 'commodities', 'forex', 'simulation'
        ]

    

    def by_market(self, market_type:str):
        """
            # Find All Datasets By Market

            market_type: ['crypto', 'stock', 'commodities', 'forex', 'simulation']
        """
        if market_type not in self.markets:
            logger.error(f"Not the correct type: {market_type} must be {self.markets}")
            return []

        _search = self.search
        _search[self.sc] = {
            "market": market_type
        }
        # print(_search.query_builder.build())
        return _search.find()
    
    def by_country(self, country:str):

        """
            # Find All Datasets By Country

        """

        if not isinstance(country, str):
            logger.error("The country is not the string.")
            return []
        _search = self.search
        _search[self.sc] = {
            "country": country,
            # "data"
        }
        return _search.find()
    
    def by_sector(self, sector:str):
        """
            # Find All Datasets By Sector

        """


        if not isinstance(sector, str):
            logger.error("The sector should be a string.")
            return []
        _search = self.search
        _search[self.sc] = {
            "sector": sector
        }
        return _search.find()
    
    def by_name(self, name:str):
        """
            # Find All Datasets By Sector

        """


        if not isinstance(name, str):
            logger.error("The sector should be a string.")
            return []
        _search = self.search
        _search["name"] = name
        return _search.find()
    

    def by_exchange(self, name:str):
        if not isinstance(name, str):
            logger.error("The sector should be a string.")
            return []
        _search = self.search
        _search[self.sc] = {
            "exchange": name
        }
        return _search.find()

    
    def multi_search(self, name=None, country=None, sector=None, market=None, exchange=None, is_exact_subcategory=False):
        """ Search with our conventional parameters for our pricing datasets """
        all_variables = {"name": name, "country": country, "sector": sector, "market": market, "exchange": exchange}
        _name = None
        _subcat_dict = {}
        for k, v in all_variables.items():
            if v is None:
                continue
            if k == "name":
                _name = v
                continue
            if k == "market":
                if v not in self.markets:
                    continue
            
            if is_exact_subcategory:
                _subcat_dict[k] = querying.text.exact(v)
            else:
                _subcat_dict[k] = v
        
        is_size = (len(_subcat_dict) == 0)
        is_name = (_name is None)
        if is_size and is_name:
            return []

        _search = self.search

        if not is_name:
            _search["name"] = name
        if not is_size:
            _search[self.sc] = _subcat_dict
        _search.processor = self.processor
        return _search.find()
    

    def build(self, name:str, abbv:str, country:str="US", sector:str="tech", market:str="stock", exchange:str="binance"):
        self['name'] = name
        self['abbreviation'] = abbv
        self['subcategories'] = {
            "market": market,
            "country": country,
            "sector": sector,
            "exchange": exchange,
        }
        return self

    # def get(self, name=None, country=None, sector=None, )

def main():
    import pandas_datareader.data as web
    # data_msft = web.DataReader('MSFT','yahoo',start='2019/9/1',end='2020/1/30').round(2)
    # data_apple = web.DataReader('AAPL','yahoo',start='2019/9/1',end='2020/1/30').round(2)
    episode_id = uuid.uuid4().hex
    jambo = Jamboree()
    jam_processor = Jamboree()
    data_hander = PriceData()
    data_hander.processor = jam_processor
    trx_tron = data_hander.build("Tron", "TRX", country="Japan", sector="oil", market="commodities", exchange="binance")
    # The episode and live parameters are probably not good for the scenario. Will probably need to switch to something else to identify data
    trx_tron.episode = episode_id
    trx_tron.live = False
    trx_tron.reset()
    

    res = trx_tron.multi_search(country="jap")
    pprint.pprint(res)


if __name__ == "__main__":
    main()


from typing import Dict, List
from copy import copy

import maya
import orjson
from loguru import logger

from jamboree.storage.databases import DatabaseConnection


class RedisDatabaseConnection(DatabaseConnection):
    """ 
        Redis Connection Notes
        ---
        We're going to end up using zadds and zscores. 
        The indexable score makes it a good target for time series. 
        The fact that it's a set makes it so there can be no duplicates added.

        In essense. Perfect for time series. 
    """

    def __init__(self) -> None:
        super().__init__()

    """
        Save commands
    """

    def _save(self, _hash: str, data: dict):
        serialized = orjson.dumps(data)
        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            push_key = f"{_hash}:list"
            self.connection.rpush(push_key, serialized)

    def save(self, query: dict, data: dict):
        if not self.helpers.validate_query(query):
            return
        timestamp = maya.now()._epoch
        _hash = self.helpers.generate_hash(query)
        query.update(data)
        query['timestamp'] = timestamp
        self._save(_hash, query)

    def _save_many(self, _hash: str, data: List[Dict]):
        serialized_list = [orjson.dumps(x) for x in data]

        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            push_key = f"{_hash}:list"
            self.connection.rpush(push_key, *serialized_list)

    def save_many(self, query, data: List[Dict] = []):
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        updated_list = [self.helpers.update_dict(query, x) for x in data]
        _hash = self.helpers.generate_hash(query)
        self._save_many(_hash, updated_list)

    """
        Update commands
    """

    def _update_single(self, _hash: str, data: dict):
        serialized = orjson.dumps(data)
        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            push_key = f"{_hash}:list"
            self.connection.rpush(push_key, serialized)

    def update_single(self, query: dict, data: dict):
        if not self.helpers.validate_query(query):
            return
        _hash = self.helpers.generate_hash(query)
        self._update_single(_hash, data)

    def _update_many(self, _hash: str, data: List[Dict] = []):
        serialized_list = [orjson.dumps(x) for x in data]

        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            push_key = f"{_hash}:list"
            self.connection.rpush(push_key, *serialized_list)

    def update_many(self, query, data: List[Dict] = []):
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        updated_list = [self.connection.update_dict(query, x) for x in data]
        _hash = self.helpers.generate_hash(query)
        self._update_many(_hash, updated_list)

    """
        # Delete Commands
        ---
        
    """

    def delete_first(self, query, details):
        pass

    def _delete(self, _hash: str, details: dict):
        phindex = self.connection.incr("placeholder_del:index")
        placeholder_hash = f"{_hash}:placeholder:{phindex}"
        placeholder_hash_del = f"{_hash}:placeholder_del:{phindex}"
        push_key = f"{_hash}:list"
        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            all_matching_redis_items = self.helpers.back_to_dict(self.connection.lrange(push_key, 0, -1))
            if isinstance(all_matching_redis_items, dict):
                """ Remove replace the current list with the empty one"""
                is_true = self.helpers.search_one(all_matching_redis_items, details)
                if is_true == False: return
                self.connection.rpush(placeholder_hash, orjson.dumps(all_matching_redis_items))
            else:
                for match in all_matching_redis_items:
                    is_true = self.helpers.search_one(match, details)
                    if is_true:
                        self.connection.rpush(placeholder_hash, orjson.dumps(match))

            self.connection.rename(push_key, placeholder_hash_del)
            self.connection.rename(placeholder_hash, push_key)

    def delete(self, query: dict, details: dict):
        if not self.helpers.validate_query(query):
            return
        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return
        self._delete(_hash, details)

    def _delete_many(self, _hash: str, data: List[Dict]):
        pass

    def delete_many(self, query, data: List[Dict]):
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        updated_list = [self.connection.update_dict(query, x) for x in data]
        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return
        self._delete_many(_hash, updated_list)

    def _delete_all(self, _hash: str):
        while self.connection.llen(_hash) > 0:
            self.connection.ltrim(_hash, 0, -99)

    def delete_all(self, query: dict, details: dict):
        if not self.helpers.validate_query(query):
            return

        _hash = self.helpers.generate_hash(query)
        self._delete_all(_hash)

    """ 
        Query commands
    """

    def query_latest(self, query: dict):
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return []

        latest_redis_items = self.helpers.back_to_dict(self.connection.lrange(f"{_hash}:list", -1, -1))
        return latest_redis_items

    def query_latest_many(self, query: dict, limit=1000):
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return []

        latest_redis_items = self.helpers.back_to_dict(self.connection.lrange(f"{_hash}:list", -limit, -1))
        return latest_redis_items

    def query_all(self, query: dict):
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return []

        latest_redis_items = self.helpers.back_to_dict(self.connection.lrange(f"{_hash}:list", 0, -1))
        return latest_redis_items

    """ Swap focused commands"""

    def _latest_many_swap(self, _hash: str, limit: int = 10):
        """ Actually do the redis operation here. """
        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            with self.connection.pipeline() as pipe:
                latest_items = []
                try:
                    push_key = f"{_hash}:list"
                    swap_key = f"{_hash}:swap"

                    pipe.watch(push_key)
                    pipe.watch(swap_key)
                    main_count = pipe.llen(push_key)
                    swap_count = pipe.llen(swap_key)

                    if main_count == 0 and swap_count == 0: raise AttributeError("Skip further queries")

                    # Determine the amount we're going to get from the main search
                    limit_swap_diff = limit - swap_count

                    main_req = 0
                    swap_req = 0

                    # ----------------------------------------------------------
                    # -------------- Get the query requirements ----------------
                    # ----------------------------------------------------------

                    # logger.info(main_count)
                    # logger.info(swap_count)
                    if limit_swap_diff < 0:
                        swap_req = limit
                    elif limit_swap_diff >= 1:
                        swap_req = swap_count
                        main_req = limit_swap_diff

                    main_latest_items = []
                    if main_req != 0:
                        main_latest_items = pipe.lrange(push_key, -main_req, -1)

                    swap_latest_items = list(reversed(pipe.lrange(swap_key, -swap_req, -1)))
                    latest_items = main_latest_items + swap_latest_items
                    # means the count of the swapped elements is less than the total limit, yet isn't empty
                    pipe.execute()
                    logger.info(latest_items)
                except Exception as e:
                    logger.error(str(e))
                finally:
                    pipe.reset()

                if len(latest_items) > 0:
                    items = self.helpers.back_to_dict(latest_items)

                    return items
                return latest_items

    def query_mix(self, query: dict, _limit: int):
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return []
        return self._latest_many_swap(_hash, _limit)

    def query_latest_swap(self, query: dict):
        if not self.helpers.validate_query(query):
            return {}

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return {}
        return self._latest_many_swap(_hash, 1)

    def _swap(self, _hash: str, limit):
        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            with self.connection.pipeline() as pipe:
                latest_items = []
                try:
                    push_key = f"{_hash}:list"
                    swap_key = f"{_hash}:swap"
                    pipe.watch(push_key)
                    pipe.watch(swap_key)

                    abs_limit = abs(limit)

                    latest_items = pipe.lrange(push_key, -(abs_limit), -1)
                    latest_items_reversed = copy(latest_items)
                    pipe.ltrim(push_key, 0, -(abs_limit + 1))
                    if len(latest_items) > 0:
                        latest_items_reversed = list(reversed(latest_items_reversed))
                        pipe.rpush(swap_key, *latest_items_reversed)

                    pipe.execute()

                except Exception as e:
                    logger.info(str(e))
                finally:
                    pipe.reset()
                if len(latest_items) > 0:
                    return self.helpers.back_to_dict(latest_items)
                return latest_items

    def swap(self, query: dict, limit: int = 100):
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return []
        return self._swap(_hash, limit)

    def swap_one(self, query):
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0: return []
        limit = 1
        return self._swap(_hash, limit)

    """ 
        Pop commands
    """

    def _pop_many(self, _hash: str, limit: int = 10):
        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            with self.connection.pipeline() as pipe:
                latest_items = []
                try:
                    push_key = f"{_hash}:list"
                    pipe.watch(push_key)
                    latest_items = pipe.lrange(push_key, -limit, -1)
                    pipe.ltrim(push_key, 0, -limit)
                    pipe.execute()

                except Exception as e:
                    pass
                finally:
                    pipe.reset()
                if len(latest_items) > 0:
                    return self.helpers.back_to_dict(latest_items)
                return latest_items

    def pop(self, query: dict, limit=1):
        if not self.helpers.validate_query(query):
            return {}
        _hash = self.helpers.generate_hash(query)
        if self.count(_hash) == 0:
            return {}
        return self._pop_many(_hash, limit=limit)

    def pop_many(self, query: dict, limit=100):
        if not self.helpers.validate_query(query):
            return []
        _hash = self.helpers.generate_hash(query)
        if self.count(_hash) == 0:
            return []
        return self._pop_many(_hash, limit=limit)

    """
        Other Functions
    """

    def _reset_count(self, query: dict, mongo_data: List[Dict]):
        """ Reset the count for the current mongodb query. We do this by adding records in mongo back into redis. """
        if len(mongo_data) == 0:
            return

        _hash = self.helpers.generate_hash(query)
        phindex = self.connection.incr("placeholder:index")
        delindex = self.connection.incr("deletion:index")
        _hash_key = f"{_hash}:list"
        _hash_placeholder = f"{_hash}:{phindex}"
        _hash_del = f"{_hash}:{delindex}"

        serialized_mongo = [orjson.dumps(mon) for mon in mongo_data]
        rlock = f"{_hash}:lock"
        with self.connection.lock(rlock):
            with self.connection.pipeline() as pipe:
                pipe.watch(_hash)
                pipe.watch(_hash_placeholder)
                pipe.watch(_hash_key)
                pipe.connection.rpush(_hash_placeholder, *serialized_mongo)
                pipe.connection.rename(_hash_key, _hash_del)
                pipe.connection.rename(_hash_placeholder, _hash_key)

                pipe.execute()

        self._delete_all(_hash_del)

    def reset(self, query, mongo_data: List[Dict]):
        if not self.helpers.validate_query(query): return
        self.pool.schedule(self._reset_count, args=(query, mongo_data))

    def count(self, _hash: str) -> int:
        _count_hash = f"{_hash}:list"
        count = self.connection.llen(_count_hash)
        return int(count)



# -*- coding: utf-8 -*-
import os
import codecs
import sys
from shutil import rmtree
from setuptools import setup, find_packages, Command


here = os.path.abspath(os.path.dirname(__file__))


with open("README.md", "r") as fh:
    long_description = fh.read()


class UploadCommand(Command):
    """Support setup.py publish."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except FileNotFoundError:
            pass
        self.status("Building Source distribution…")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))
        self.status("Uploading the package to PyPi via Twine…")
        os.system("twine upload dist/*")
        sys.exit()



setup(
    name="jamboree",
    version="0.9.3",
    author="Kevin Hill",
    author_email="kah.kevin.hill@gmail.com",
    description="A multi-layer event sourcing and general data library. SQL, Search, Event Sourcing, and File/Model storage combined into one.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    py_modules=["jamboree"],
    install_requires=[
        "numpy",
        "scipy",
        'maya',
        "pandas-datareader",
        "json-tricks",
        "ujson",
        'gym',
        "orjson",
        "pebble",
        "cytoolz",
        "loguru",
        'redisearch',
        "lz4",
        "anycache",
        "hiredis",
        "eliot",
        "eliot-tree",
        "matplotlib",
        "dask",
        "pandas",
        "sklearn",
        "crayons",
        "creme",
        'pydantic',
        "yfinance",
        "version_query",
        "pandas",
        "cerberus",
        "addict",
    ],
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    cmdclass={"upload": UploadCommand},
)



from cerberus import Validator
from redisearch import TextField, NumericField, TagField, GeoField
# from jamboree.utils.support.search import filtration_schemas

_global_validator = Validator(require_all=True, allow_unknown=True)
# _filtration_schemas = filtration_schemas()

class Geo(type):
    """ A geolocational type for """
    def __call__(cls):
        return cls.__new__(cls)
    def __repr__(self):
        return "GEO"
    
    def __str__(self):
        return "GEO"
    

def is_nested(d):
    return any(isinstance(i,dict) for i in d.values())

def is_gen_type(item, _type):
    try:
        return isinstance(item, _type) or issubclass(item, _type) or item == _type
    except:
        return False

def name_match(item:str, name:str):
    return item.lower() == name.lower()


def is_generic(_k):
    return _k in [str, float, int, list, bool]


def is_geo(k) -> bool:
    if is_gen_type(k, Geo):
        return True
    
    if is_gen_type(k, str):
        if name_match(k, "geo"):
            return True
    return False

def to_str(i):
    """Converts the item to a string version of it"""
    if i == bool:
        # This will be text that we'll force exact queries on
        return "BOOL"
    elif i == float or i == int:
        return "NUMERIC"
    elif i == str:
        return "TEXT"
    elif i == list:
        return "TAG"


def to_field(k, v):
    if v == "BOOL":
        return TextField(k, sortable=True)
    elif v == "NUMERIC":
        return NumericField(k, sortable=True)
    elif v == "TEXT":
        return TextField(k)
    elif v == "TAG":
        return TagField(k)
    else:
        return GeoField(k)



"""
    Dictionary Validation
"""



def is_valid_geo(_dict:dict):
    """ That we have the appropiate values """
    schema = {
        "long": {"type": "number"},
        "lat": {"type": "number"},
        "distance": {"type": "number", "required":False},
        "metric": {"type": "string", "allowed": ["m","km","mi","ft"], "required":False}
    }
    return _global_validator.validate(_dict, schema)

def is_valid_bool(_dict:dict):
    """ That we have the appropiate values to create a query function for a boolean """
    schema = {
        "toggle": {"type": "boolean"},
    }
    return _global_validator.validate(_dict, schema)

def is_valid_numeric(_dict:dict):
    """ That we have the appropiate values to do a numeric query """
    schema = {
        "operation": {"type": "string", "allowed": ['greater', 'lesser', 'between', 'exact']},
        "upper": {"type": "number"}, 
        "lower": {"type": "number"}
    }
    return _global_validator.validate(_dict, schema)

def is_valid_tags(_dict:dict):
    schema = {
        "operation": {"type": "string", "allowed": ['and', 'or']},
        "tags": {"type": "list", "schema": {"type": "string"}}, 
    }
    return _global_validator.validate(_dict, schema)

def is_valid_text(_dict:dict):
    schema = {
        "term": {"type": "string"},
        "is_exact": {"type": "boolean", "required":False}, 
    }
    return _global_validator.validate(_dict, schema)

def is_queryable_dict(_dict:dict):
    """ """
    schema = {
        "type": {
            "type": "string", 
            "allowed": ["GEO", "TEXT", "BOOL", "NUMERIC", "TAG"]
        },
        "is_filter": {
            "type": "boolean"
        },
        "values": {
            "type": "dict"
        }
    }
    return _global_validator.validate(_dict, schema)

# Specific queryable information


def main():
    _search_item = {
        "type": "GEO",
        "is_filter": False,
        "values": {
            "long": 33,
            "lat": -10,
            "distance": 1,
            "metric": "km"
        }
    }

    _numeric_search_item = {
        "operation": "between",
        "upper": 0,
        "lower": 0
    }

    _bool_search_values = {
        "toggle": True
    }

    print(is_queryable_dict(_search_item))
    print(is_valid_numeric(_numeric_search_item))
    print(is_valid_bool(_bool_search_values))


if __name__ == "__main__":
    main()


import uuid

import maya
import pandas as pd
import ujson
from loguru import logger

from jamboree import Jamboree
from jamboree.handlers.abstracted.search import MetadataSearchHandler
from jamboree.handlers.complex.meta import MetaHandler
from jamboree.handlers.default import Access, DBHandler, TimeHandler
from jamboree.handlers.processors import (DataProcessorsAbstract,
                                          DynamicResample)
from jamboree.utils import omit
from jamboree.utils.support.search import querying


class DataHandler(Access):
    """ 
        # DATA HANDLER
        ---

        The data handler does the following:
        
        1. Accepts DataFrame input with or without time index
        2. Converts and saves the data
        3. Attaches to Metadata (needs to have a metadata handler to ensure effective usage of information)
        4. Connects to a MultiDataHandler
            * The MultiDataHandler
        5. Adds autoresampling functionality.

    """

    def __init__(self):
        super().__init__()
        self.entity = "data"

        self.required = {
            "name": str,
            "category": str,
            "subcategories": dict,
            "metatype": str,
            "submetatype": str,
            "abbreviation": str,
        }
        self._time: TimeHandler = TimeHandler()
        self._meta: MetaHandler = MetaHandler()
        self._metasearch = MetadataSearchHandler()
        self._episode = uuid.uuid4().hex
        self._is_live = False
        self._preprocessor: DataProcessorsAbstract = DynamicResample("data")
        self.is_event = False  # use to make sure there's absolutely no duplicate data
        self.metaid = ""
        self["metatype"] = self.entity

        self.is_robust = False

    @property
    def episode(self) -> str:
        return self._episode

    @episode.setter
    def episode(self, _episode: str):
        self._episode = _episode

    @property
    def live(self) -> bool:
        return self._is_live

    @live.setter
    def live(self, _live: bool):
        self._is_live = _live

    @property
    def time(self) -> "TimeHandler":
        self._time.processor = self.processor
        self._time["episode"] = self.episode
        self._time["live"] = self.live
        return self._time

    @time.setter
    def time(self, _time: "TimeHandler"):
        self._time = _time

    @property
    def metadata(self):
        self._meta.processor = self.processor
        self._meta["name"] = self.name
        self._meta["category"] = self.category
        self._meta["subcategories"] = self.subcategories
        self._meta["metatype"] = self.entity
        self._meta["submetatype"] = self.submetatype
        self._meta["abbreviation"] = self.abbreviation
        return self._meta

    @property
    def search(self):
        self._metasearch.reset()
        self._metasearch["category"] = querying.text.exact(self.category)
        self._metasearch["metatype"] = querying.text.exact(self.entity)
        self._metasearch["submetatype"] = querying.text.exact(self.submetatype)

        self._metasearch.processor = self.processor
        return self._metasearch

    @property
    def is_next(self) -> bool:
        """ A boolean that determines if there's anything next """

        next_data = self.closest_head()
        next_keys = list(next_data.keys())
        if len(next_keys) == 0:
            return False
        return True

    @property
    def preprocessor(self) -> DataProcessorsAbstract:
        return self._preprocessor

    @preprocessor.setter
    def preprocessor(self, _preprocessor: DataProcessorsAbstract):
        self._preprocessor = _preprocessor

    def _timestamp_resample_and_drop(self, frame: pd.DataFrame, resample_size="D"):
        timestamps = pd.to_datetime(frame.time, unit="s")
        frame.set_index(timestamps, inplace=True)
        frame = frame.drop(
            columns=["timestamp", "type", "subcategories", "category", "time"]
        )
        frame = frame.resample(resample_size).mean()
        frame = frame.fillna(method="ffill")
        return frame
    def store_time_df(self, dataframe: pd.DataFrame, is_bar=False):
        """ 
            Breaks a dataframe into parts then stores them into redis. 
            Use to handle time series data. Dataframe must have time index
        """
        storable_list = self.main_helper.convert_dataframe_to_storable_item_list(
            dataframe
        )
        if is_bar == True:
            storable_list = self.main_helper.standardize_outputs(storable_list)
        self.save_many(storable_list)

    def add_now(self, data_dict: dict, is_bar=False):
        """ Add information to the current dataset"""
        data_dict_list = [data_dict]
        if is_bar == True:
            data_dict_list = self.main_helper.standardize_outputs(data_dict_list)
        self.save_many(data_dict_list)

    def dataframe_from_head(self):
        """ Get a dataframe between a head and tail. Resample according to our settings"""

        head = self.time.head
        tail = self.time.tail
        values = self.in_between(tail, head, ar="relative")
        frame = pd.DataFrame(values)
        frame = self._timestamp_resample_and_drop(frame)
        return frame
    

    def dataframe_all(self):
        """ Get a dataframe between a head and tail. Resample according to our settings"""

        values = self.query_all()
        frame = pd.DataFrame(values)
        frame = self._timestamp_resample_and_drop(frame)
        return frame

    def dataframe_from_dynamic_peak(self, n_head: int = 1, n_tail=100000):
        """ Get a dataframe between a head and tail. Resample according to our settings"""

        head = self.time.peak_back_num(n_head)
        tail = self.time.peak_back_num_tail(n_tail)
        values = self.in_between(tail, head, ar="relative")
        frame = pd.DataFrame(values)
        frame = self._timestamp_resample_and_drop(frame)
        return frame

    def dataframe_from_last(self):
        """ Get a dataframe with all of the last information. Resample according to our settings"""

        values = self.many(ar="relative")
        frame = pd.DataFrame(values)
        frame = self._timestamp_resample_and_drop(frame)
        return frame

    def closest_head(self, is_robust=False):
        """ Get the closest information at the given head. Otherwise get the latest information"""
        head = self.time.head
        count = self.count()
        closest = self.last_by(head, ar="relative")
        if len(closest) == 0:
            if (is_robust or self.is_robust) and count > 0:
                last = self.last(ar="relative")
                last.pop("name", None)
                last.pop("mtype", None)
                last.pop("category", None)
                last.pop("subcategories", None)
                last.pop("type", None)
                return last
            return {}
        closest.pop("name", None)
        closest.pop("category", None)
        closest.pop("subcategories", None)
        closest.pop("type", None)
        return closest

    def closest_peakback_by(self, n: int = 1, is_robust=False):
        """ Get the closest information at the given head. Otherwise get the latest information"""
        head = self.time.peak_back_num(n)
        count = self.count()
        closest = self.last_by(head, ar="relative")
        if len(closest) == 0:
            if is_robust or self.is_robust:
                if count > 0:
                    last = self.last(ar="relative")
                    last.pop("name", None)
                    last.pop("mtype", None)
                    last.pop("category", None)
                    last.pop("subcategories", None)
                    last.pop("type", None)
                    return last
                return {}
        closest.pop("name", None)
        closest.pop("category", None)
        closest.pop("subcategories", None)
        closest.pop("type", None)
        return closest

    def closest_head_omitted(self, is_robust=False):
        """ Get the closest information at the given head. Otherwise get the latest information"""
        head = self.time.head
        count = self.count()
        closest = self.last_by(head, ar="relative")
        omit_list = [
            "metatype",
            "submetatype",
            "abbreviation",
            "mtype",
            "category",
            "subcategories",
            "name",
            "type",
        ]
        if len(closest) == 0:
            if (is_robust or self.is_robust) and count > 0:
                last = self.last(ar="relative")
                last = omit(omit_list, last)
                return last
            return {}

        closest = omit(omit_list, closest)
        return closest

    def closest_peakback_by_omitted(self, n: int = 1, is_robust=False):
        """ Get the closest information at the given head. Otherwise get the latest information"""
        head = self.time.peak_back_num(n)
        count = self.count()
        closest = self.last_by(head, ar="relative")
        omit_list = [
            "metatype",
            "submetatype",
            "abbreviation",
            "mtype",
            "category",
            "subcategories",
            "name",
            "type",
        ]
        if len(closest) == 0:
            if is_robust or self.is_robust:
                if count > 0:
                    last = self.last(ar="relative")
                    last = omit(omit_list, last)
                    return last
                return {}
        closest = omit(omit_list, closest)
        return closest

    def previous_head(self):
        """ Get the closest information at the given head"""
        head = self.time.peak_back()
        closest = self.last_by(head, ar="relative")
        return closest

    def reset(self):
        """ Reset the data we're querying for. """

        # Update. Get the highest and lowest score for the current dataset.
        self.time.reset()
        self.metaid = self.metadata.reset()
        return self.metaid

    def pick(self, _id: str) -> None:
        """Get the metadata to populate the current fields and loads the data.

        Args:
            _id (str): The metadata identifier.

        Raises:
            AttributeError: Throws an attribute error if metadata couldn't be found.
        """
        current_search = self.search
        item = current_search.pick(_id)

        if item is None:
            raise AttributeError("Data source currently doesn't exist.")

        self["name"] = item["name"]
        self["category"] = item["category"]
        self["subcategories"] = item["subcategories"]
        self["metatype"] = item["metatype"]
        self.entity = item["metatype"]
        self["submetatype"] = item["submetatype"]
        self["abbreviation"] = item["abbreviation"]
        self.reset()

    def __str__(self) -> str:
        name: str = self["name"]
        category: str = self["category"]
        subcategories: dict = self["subcategories"]
        metatype: str = self["metatype"]
        submetatype: str = self["submetatype"]
        abbreviation: str = self["abbreviation"]

        jscat = self.main_helper.generate_hash(subcategories)
        return f"{name}:{category}:{jscat}:{metatype}:{submetatype}:{abbreviation}"


if __name__ == "__main__":
    import pandas_datareader.data as web

    data_msft = web.DataReader(
        "MSFT", "yahoo", start="2010/1/1", end="2020/1/30"
    ).round(2)
    data_apple = web.DataReader(
        "AAPL", "yahoo", start="2010/1/1", end="2020/1/30"
    ).round(2)
    # print(data_apple)
    episode_id = uuid.uuid4().hex
    jambo = Jamboree()
    data_hander = DataHandler()
    data_hander.event = jambo
    data_hander.processor = jambo
    # The episode and live parameters are probably not good for the scenario. Will probably need to switch to something else to identify data
    data_hander.episode = episode_id
    data_hander.live = False
    data_hander["category"] = "markets"
    data_hander["subcategories"] = {
        "market": "stock",
        "country": "US",
        "sector": "techologyyyyyyyy",
    }
    data_hander["name"] = "MSFT"
    data_hander["submetatype"] = "POOP"
    data_hander["abbreviation"] = "MSFT"
    data_hander.reset()
    data_hander.store_time_df(data_msft, is_bar=True)
    data_hander["name"] = "AAPL"
    data_hander.store_time_df(data_apple, is_bar=True)
    data_hander.reset()

    data_hander.time.head = maya.now().subtract(weeks=200, hours=14)._epoch
    data_hander.time.change_stepsize(microseconds=0, days=1, hours=0)
    data_hander.time.change_lookback(microseconds=0, weeks=4, hours=0)

    while data_hander.is_next:

        logger.success(data_hander.name)
        logger.warning(data_hander.category)
        logger.error(data_hander.subcategories)
        logger.debug(data_hander.abbreviation)
        logger.success(data_hander.submetatype)
        
        print()
        print()
        logger.success(data_hander.time.head)
        logger.info(data_hander.time.head)
        logger.success(data_hander.time.head)
        print()
        print()
        data_hander.time.step()



import copy
from abc import ABC, ABCMeta
from typing import Dict, Any, List


class BaseHandler(object, metaclass=ABCMeta):
    """ 
        A way to handle reads and writes consistently without having to write every single variable:
    """

    def __init__(self):
        pass

    def check(self):
        raise NotImplementedError

    def save(self, data: dict):
        raise NotImplementedError

    def _bulk_save(self, query: dict, data: list):
        raise NotImplementedError

    def _get_many(self):
        raise NotImplementedError

    def last(self):
        raise NotImplementedError

    def many(self, limit: int = 100):
        raise NotImplementedError

    def save_many(self, query: dict, data: list):
        raise NotImplementedError

    def pop_multiple(self, query, _limit: int = 1):
        raise NotImplementedError

    def swap(self, query, alt: dict = {}):
        """ Swap betwen the first and last item """
        raise NotImplementedError

    def query_mix(self, query: dict, alt: dict = {}):
        raise NotImplementedError


class BaseFileHandler(object, metaclass=ABCMeta):
    """ 
        A way to handle reads and writes consistently without having to write every single variable:
    """

    def __init__(self):
        pass

    def check(self):
        raise NotImplementedError

    def save(self, data: dict):
        raise NotImplementedError
    
    def save_version(self, query:dict, data):
        pass

    def last(self):
        raise NotImplementedError

    def many(self, limit: int = 100):
        raise NotImplementedError

    def save_many(self, query: dict, data: list):
        raise NotImplementedError

    def delete(self, query:dict):
        raise NotImplementedError
    
    def delete_version(self, query:dict, version:str):
        raise NotImplementedError


import functools
import weakref

from cytoolz import keyfilter


def omit(blacklist, d):
    return keyfilter(lambda k: k not in blacklist, d)

def memoized_method(*lru_args, **lru_kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            self_weak = weakref.ref(self)
            @functools.wraps(func)
            @functools.lru_cache(*lru_args, **lru_kwargs)
            def cached_method(*args, **kwargs):
                return func(self_weak(), *args, **kwargs)
            setattr(self, func.__name__, cached_method)
            return cached_method(*args, **kwargs)
        return wrapped_func
    return decorator



import time
import uuid
from typing import Optional

import maya
from jamboree import Jamboree
from jamboree.handlers.abstracted.search import MetadataSearchHandler
from jamboree.handlers.default import DBHandler, Access
from loguru import logger


class MetaHandler(Access):
    """ 
        # MetaDataHandler
        --- 
        Metadata is "data that provides information about other data".
        
        The MetaHandler is a way to interact with metadata on each data source we have. 
        
        
        It should be used with both the DataHandler an MultDataHandler. As well as any other form of common data we're looking for as well. 
        It would be there in the event that we would want to figure out properties of data without being forced to directly open the data.
        It should also give us the capacity to search for various bits of information (redis_search) in the near future.

        Some usecases of the metadata could include:

        1. Knowing the type of data we're looking at given some information.
            - Time-series
            - Machine Learning Model
            - Network/Graph Data
            - Events
            - Log Data
            - Meta Record
                - A metarecord is a json representative to a complex datatype. 
        2. Times the data was initiatied
        3. Time the data was last modified
            - Modifications can be as simple as: 
                - Adding a new ticker or bar for price information
                - Partial-Fitting a machine learning model
                - Adjusting a weight to a variable
        4. Getting the number of records for a given piece of information
            - Very useful if we're trying to plan around how much we're going to do for a piece of information
        5. Determining if such data exist
            - We would simply create a complex hash function that's pulled from all dbhandlers representing that data type. 
        6. Start and End Time for a given set of records
        7. Location information
            - There can be different location information for each piece of information.
            - Examples:
                - Image weather data
                - Market location data
                - Social interaction location data
                - Login, logout location data
            - Creating something flexible for this would probably be a good idea. 
    """
    def __init__(self):
        super().__init__()
        self.entity = "metadata"
        self.required = {
            "name": str,
            "category": str,
            "metatype": str,
            "submetatype": str,
            "abbreviation": str,
            "subcategories": dict
        }
        self._search = MetadataSearchHandler()
        self._settings = {}
        self.is_auto = False
        self.description: Optional[str] = None

    @property
    def search(self):
        metatype = self.metatype
        submetatype = self.submetatype
        self._search.entity = self.entity
        self._search['metatype'] = {
            "type": "TEXT",
            "is_filter": True,
            "values": {
                "is_exact": True,
                "term": metatype
            }
        }
        self._search['submetatype'] = {
            "type": "TEXT",
            "is_filter": True,
            "values": {
                "is_exact": True,
                "term": submetatype
            }
        }
        self._search['name']            = self.name
        self._search['category']        = self.category
        self._search['subcategories']   = self.subcategories
        self._search['abbreviation']    = self.abbreviation
        self._search.processor          = self.processor
        return self._search

    def reset(self):
        self.check()
        qo = self.setup_query()
        qo.pop("mtype", None)
        if self.description is not None:
            qo['description'] = self.description
        
        return self.search.Create(allow_duplicates=False,
                                  no_overwrite_must_have=True,
                                  **qo)


if __name__ == "__main__":
    jambo = Jamboree()
    metahandler = MetaHandler()
    metahandler.event = jambo
    metahandler.reset()



from .validation import is_nested, is_gen_type, name_match, is_generic, is_geo, to_str, to_field, is_queryable_dict
from .validation import is_valid_geo, is_valid_bool, is_valid_numeric, is_valid_tags, is_valid_text
from .builders import InsertBuilder, QueryBuilder
from .core import BaseSearchHandlerSupport


import base64
import random
from abc import ABC
from copy import copy
from multiprocessing import cpu_count
from typing import List

import maya
import orjson
import ujson
from loguru import logger
from pebble.pool import ThreadPool
from redis import Redis

from funtime import Store


class EventProcessor(ABC):
    def save(self, query: dict, data: dict):
        raise NotImplementedError

    def get_latest(self, query):
        raise NotImplementedError

    def get_latest_many(self, query, limit=1000):
        raise NotImplementedError

    def save_many(self, query: dict, data: List[dict]):
        raise NotImplementedError

    def count(self, query: dict):
        raise NotImplementedError

    def remove_first(self, query: dict):
        raise NotImplementedError

    def pop_multiple(self, query: dict, limit: int):
        raise NotImplementedError

    def _bulk_save(self, query: dict, data: list):
        raise NotImplementedError

    def multi_swap(self, query: dict, limit: int = 100):
        raise NotImplementedError

    def query_mix(self, query: dict, limit: int = 100):
        raise NotImplementedError

    def get_latest_many_swap(self, query: dict, limit: int):
        raise NotImplementedError


class Jamboree(EventProcessor):
    """Adds and retrieves events at extremely fast speeds. Use to handle portfolio and trade information quickly."""

    def __init__(self, mongodb_host="localhost", redis_host="localhost", redis_port=6379):
        self.redis = Redis(redis_host, port=redis_port)
        self.store = Store(mongodb_host).create_lib('events').get_store()['events']
        self.pool = ThreadPool(max_workers=cpu_count() * 4)

    def _validate_query(self, query: dict):
        """ Validates a query. Must have `type` and a second identifier at least"""
        if 'type' not in query:
            return False
        if not isinstance(query['type'], str):
            return False
        if len(query) < 2:
            return False
        return True

    def _generate_hash(self, query: dict):
        _hash = ujson.dumps(query, sort_keys=True)
        _hash = base64.b64encode(str.encode(_hash))
        _hash = _hash.decode('utf-8')
        return _hash

    def _check_redis_for_prior(self, _hash: str) -> bool:
        """ Checks to see if any """
        prior_length = self.redis.llen(_hash)
        if prior_length == 0:
            return False
        return True

    def _update_dict(self, query: dict, data: dict):
        query = copy(query)
        timestamp = maya.now()._epoch
        query['timestamp'] = timestamp
        data.update(query)
        return data

    def _update_dict_no_timestamp(self, query: dict, data: dict):
        query = copy(query)
        data = copy(data)
        data.update(query)
        data.pop("timestamp", None)
        return data

    def _omit_timestamp(self, data: dict):
        """ Removes timestamp if it exists. Use it to create a copied version of a dictionary to be saved in the
        duplicate list """
        _data = copy(data)
        _data.pop("timestamp", None)
        return _data

    def back_to_dict(self, list_of_serialized: list):
        deserialized = []
        if len(list_of_serialized) == 1:
            return orjson.loads(list_of_serialized[0])

        for i in list_of_serialized:
            deserialized.append(orjson.loads(i))
        return deserialized

    def _save(self, query: dict, data: dict):
        """
            - Given a type (data entity), data and a epoch for time (utc time only), save the data in both redis and mongo. 
            - Does it in a background process. Use with add event.
            - We save the information both in mongodb and redis. We assume there's many of each collection. 
                - We find a specific collection using the query.
        """
        if self._validate_query(query) == False:
            # Log a warning here instead
            return
        timestamp = maya.now()._epoch
        _hash = self._generate_hash(query)
        # Now time to update the system
        query.update(data)
        query['timestamp'] = timestamp

        self._save_redis(_hash, query)
        self.pool.schedule(self._save_mongo, args=(query))

    """
        RESET FUNCTIONS
    """

    def _reset_count(self, query: dict):
        """ Reset the count for the current mongodb query. We do this by adding records in mongo back into redis. """
        _hash = self._generate_hash(query)
        phindex = self.redis.incr("placeholder:index")
        delindex = self.redis.incr("deletion:index")
        _hash_key = f"{_hash}:list"
        _hash_placeholder = f"{_hash}:{phindex}"
        _hash_del = f"{_hash}:{delindex}"

        # self.redis.rename(_hash_key, _hash_rename)
        mongo_data = list(self.store.query(query))
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            # placeholder key
            for md in mongo_data:
                self.redis.rpush(_hash_placeholder, orjson.dumps(md))

            self.redis.rename(_hash_key, _hash_del)
            self.redis.rename(_hash_placeholder, _hash_key)

        self._concurrent_delete_list(_hash_del)

    def reset(self, query: dict):
        """ Resets all of the variables """
        if self._validate_query(query) == False:
            # Log a warning here instead
            return
        self.pool.schedule(self._reset_count, args=(query))

    """
        DELETES FUNCTIONS
    """

    def _concurrent_delete_list(self, key):
        while self.redis.llen(key) > 0:
            self.redis.ltrim(key, 0, -99)

    def _concurrent_delete_many(self, query: dict, details: dict):
        # combine query and details
        query.update(details)
        self.store.delete_many(query)

    def _remove(self, query: dict, details: dict):
        """ Use to both remove items from redis and mongo. Add it when you need it."""

        """ 
            Removes the given query information from the database. 
            It's a heavy computation on redis, as it'll require searching an entire list.
            
        """
        # Deletes from mongo concurrently
        self.pool.schedule(self._concurrent_delete_many, args=(query, details))
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        phindex = self.redis.incr("placeholder_del:index")
        placeholder_hash = f"{_hash}:placeholder:{phindex}"
        placeholder_hash_del = f"{_hash}:placeholder_del:{phindex}"
        push_key = f"{_hash}:list"
        rlock = f"{_hash}:lock"

        with self.redis.lock(rlock):
            all_matching_redis_items = self.back_to_dict(self.redis.lrange(push_key, 0, -1))
            if isinstance(all_matching_redis_items, dict):
                """ Remove replace the current list with the empty one"""
                is_true = self._search_one(all_matching_redis_items, details)
                if is_true == False: return
                self.redis.rpush(placeholder_hash, orjson.dumps(all_matching_redis_items))
            else:
                for match in all_matching_redis_items:
                    is_true = self._search_one(match, details)
                    if is_true:
                        self.redis.rpush(placeholder_hash, orjson.dumps(match))

            self.redis.rename(push_key, placeholder_hash_del)
            self.redis.rename(placeholder_hash, push_key)

        self.pool.schedule(self._concurrent_delete_list, args=(placeholder_hash_del))
        # Delete while unlocked.
        # self._concurrent_delete_list(placeholder_hash_del)

    def _remove_first_redis(self, _hash, query: dict):
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            push_key = f"{_hash}:list"
            self.redis.rpop(push_key)

    def remove_first(self, query: dict):
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)

        if count == 0:
            return

        self._remove_first_redis(_hash, query)

    """ 
        SAVE FUNCTIONS
    """

    def save(self, query: dict, data: dict):
        self._save(query, data)

    def save_many(self, query: dict, data: List[dict]):
        if self._validate_query(query) == False:
            # Log a warning here instead
            return

        if len(data) == 0:
            return

        for item in data:
            self._save(query, item)

    def _bulk_upsert_redis(self, query, data):
        logger.info("Default retcon redis")
        if self._validate_query(query) == False or len(data) == 0:
            # Log a warning here instead
            return {}
        updated_list_no_timestamp = [self._update_dict_no_timestamp(query, x) for x in data]

        _hash = self._generate_hash(query)
        return {
            "hash": _hash,
            "updated": updated_list_no_timestamp
        }

    def bulk_upsert_redis(self, query, data):
        return self._bulk_upsert_redis(query, data)

    def _bulk_save(self, query, data: list):
        """ Bulk adds a list to redis."""
        if self._validate_query(query) == False or len(data) == 0:
            # Log a warning here instead
            return

        updated_list = [self._update_dict(query, x) for x in data]
        updated_list_no_timestamp = [self._update_dict_no_timestamp(query, x) for x in data]
        _hash = self._generate_hash(query)

        self._bulk_save_redis(_hash, updated_list)
        self.pool.schedule(self._bulk_save_mongo, args=(query, updated_list))

    def _save_redis(self, _hash: str, data: dict):

        serialized = orjson.dumps(data)
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            push_key = f"{_hash}:list"
            self.redis.rpush(push_key, serialized)

    def _bulk_save_redis(self, _hash: str, data: list):
        serialized_list = [orjson.dumps(x) for x in data]

        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            push_key = f"{_hash}:list"
            self.redis.rpush(push_key, *serialized_list)

    def _bulk_save_mongo(self, query: dict, data: list):
        if len(data) == 0:
            return

        first_item = data[0]
        first_item.update(query)
        updated_list = [self._update_dict(query, x) for x in data]
        self.store.bulk_upsert(updated_list, _column_first=first_item.keys(), _in=['timestamp'])

    def _save_mongo(self, data):
        self.store.store(data)

    def _pop_redis_multiple(self, _hash, limit: int):
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            with self.redis.pipeline() as pipe:
                latest_items = []
                try:
                    push_key = f"{_hash}:list"
                    pipe.watch(push_key)
                    latest_items = pipe.lrange(push_key, -limit, -1)
                    pipe.ltrim(push_key, 0, -limit)
                    pipe.execute()

                except Exception as e:
                    pass
                finally:
                    pipe.reset()
                if len(latest_items) > 0:
                    return self.back_to_dict(latest_items)
                return latest_items

    def pop_multiple(self, query, limit: int = 1):
        """ Get multiple items """
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0:
            return []
        return self._pop_redis_multiple(_hash, limit)

    """
        Public Query Functions
    """

    def query_direct(self, query):
        """ Queries from mongodb directly. Used to search extremely large queries. """
        latest_items = list(self.store.query_latest(query))
        return latest_items

    def query_direct_latest(self, query):
        """ Queries from mongodb directly. Used to search extremely large queries. """
        latest_items = list(self.store.query_latest(query))
        if len(latest_items) > 0:
            return latest_items[0]
        return {}

    def get_latest(self, query):
        """ Gets the latest query"""
        # Add a conditional time lock
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count > 0:
            return self.back_to_dict(self.redis.lrange(f"{_hash}:list", -1, -1))
        # Mongo, slowdown
        latest_items = list(self.store.query_latest(query))
        if len(latest_items) > 0:
            return latest_items[0]
        return {}

    def get_latest_many(self, query: dict, limit=1000):

        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0: return []

        latest_redis_items = self.back_to_dict(self.redis.lrange(f"{_hash}:list", -limit, -1))
        # TODO: Get mongo tasks here
        # How will this work now?
        rlen = len(latest_redis_items)
        if rlen == 0:
            query["limit"] = limit
            latest_items = list(self.store.query_latest(query))
            self.reset(query)
            return latest_items

        return latest_redis_items

    """
        SEARCH ONE FUNCTIONS
    """

    def _search_one(self, item: dict, query: dict):
        all_bools = []
        for q in query:
            if q in item:
                if query[q] == query[q]:
                    all_bools.append(True)
                else:
                    all_bools.append(False)
            else:
                all_bools.append(False)
        return any(all_bools)

    def _latest_many_swap(self, _hash: str, limit: int = 10):
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            with self.redis.pipeline() as pipe:
                latest_items = []
                try:
                    swap_key = f"{_hash}:swap"
                    pipe.watch(swap_key)

                    abs_limit = abs(limit)

                    latest_items = pipe.lrange(swap_key, -abs_limit, -1)
                    latest_items = list(reversed(latest_items))
                    pipe.execute()
                except Exception as e:
                    logger.error(str(e))
                finally:
                    pipe.reset()
                if len(latest_items) > 0:
                    return self.back_to_dict(latest_items)
                return latest_items

    def get_latest_many_swap(self, query: dict, _limit: int):
        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0: return []
        return self._latest_many_swap(_hash, _limit)

    def _get_count(self, _hash: str, query: dict):
        # Checks to see if a count already exist in redis, if not, check for a count in mongo.
        _count_hash = f"{_hash}:list"
        count = self.redis.llen(_count_hash)
        if count is not None:
            return count

        records = list(self.store.query(query))
        record_len = len(records)
        return record_len

    def count(self, query):
        """ """
        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        return count

    def _multi_swap(self, _hash: str, limit=100):
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            with self.redis.pipeline() as pipe:
                latest_items = []
                try:
                    push_key = f"{_hash}:list"
                    swap_key = f"{_hash}:swap"
                    pipe.watch(push_key)
                    pipe.watch(swap_key)

                    abs_limit = abs(limit)

                    latest_items = pipe.lrange(push_key, -(abs_limit), -1)
                    latest_items_reversed = copy(latest_items)
                    pipe.ltrim(push_key, 0, -(abs_limit + 1))
                    if len(latest_items) > 0:
                        latest_items_reversed = list(reversed(latest_items_reversed))
                        pipe.rpush(swap_key, *latest_items_reversed)

                    pipe.execute()

                except Exception as e:
                    logger.info(str(e))
                finally:
                    pipe.reset()
                if len(latest_items) > 0:
                    return self.back_to_dict(latest_items)
                return latest_items

    def multi_swap(self, query: dict, limit=100):
        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0: return
        return self._multi_swap(_hash, limit=limit)

    def _query_mix(self, _hash: str, limit: int = 10) -> List:
        """ Actually do the redis operation here. """
        rlock = f"{_hash}:lock"
        with self.redis.lock(rlock):
            with self.redis.pipeline() as pipe:
                latest_items = []
                try:
                    push_key = f"{_hash}:list"
                    swap_key = f"{_hash}:swap"

                    pipe.watch(push_key)
                    pipe.watch(swap_key)
                    main_count = pipe.llen(push_key)
                    swap_count = pipe.llen(swap_key)

                    if main_count == 0 and swap_count == 0: raise AttributeError("Skip further queries")

                    # Determine the amount we're going to get from the main search
                    limit_swap_diff = limit - swap_count

                    main_req = 0
                    swap_req = 0

                    # ----------------------------------------------------------
                    # -------------- Get the query requirements ----------------
                    # ----------------------------------------------------------

                    if limit_swap_diff < 0:
                        swap_req = limit
                    elif limit_swap_diff >= 1:
                        swap_req = swap_count
                        main_req = limit_swap_diff

                    main_latest_items = []
                    if main_req != 0:
                        main_latest_items = pipe.lrange(push_key, -main_req, -1)

                    swap_latest_items = list(reversed(pipe.lrange(swap_key, -swap_req, -1)))

                    latest_items = main_latest_items + swap_latest_items
                    # means the count of the swapped elements is less than the total limit, yet isn't empty
                    pipe.execute()

                except Exception as e:
                    logger.error(str(e))
                finally:
                    pipe.reset()
                if len(latest_items) > 0:
                    return self.back_to_dict(latest_items)
                return latest_items

    def query_mix(self, query: dict, limit=100) -> List:
        """ Get information from both the `:swap` and real list. Empty list for now. """
        if self._validate_query(query) == False: return []
        _hash = self._generate_hash(query)
        count = self._get_count(_hash, query)
        if count == 0: return []
        return self._query_mix(_hash, limit=limit)



from abc import ABC
from typing import List

class SearchProcessor(ABC):
    """ All of the common search queries will go here """
    def search(self):
        pass


    def search_specific(self):
        pass

    def search_all(self):
        """ Search all of the tags"""
        pass


from typing import List
from addict import Dict as ADDICT
from jamboree.middleware.procedures import ProcedureManagement, ModelProcedureAbstract
# from jamboree.middleware.procedures import (
#     CremeProcedure,
#     ModelProcedureAbstract,
#     SklearnProcedure,
#     TFKerasProcedure,
#     TorchProcedure,
# )
from loguru import logger


# class ModelProcedureManagement(ProcedureManagement):
#     def __init__(self):
#         super().__init__()
#         self.required_attributes = ["model_type", "online"]
        
#         # Don't forget to look at the 

#     @property
#     def allowed(self) -> List[str]:
#         return ["sklearn", "torch", "keras", "creme"]

    
#     def access(self, key):
#         self.check_allowed(key)
#         spec = {
#             "sklearn": SklearnProcedure(),
#             "torch": TorchProcedure(),
#             "keras": TFKerasProcedure(),
#             "creme": CremeProcedure(),
#         }.get(key)
#         return spec


# if __name__ == "__main__":
#     modelmgt = ModelProcedureManagement()
#     print(modelmgt.access("sklearn"))



import uuid
import maya

from typing import Optional
from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.handlers.default.search import BaseSearchHandler

from loguru import logger

class ParameterizedSearch(BaseSearchHandler):
    """ 
        An abstract over the base search handler. 
        
        Use to avoid using the normal formatting. 

        Example:

        Normally you'd have to use the following:

        ::
            >>> search['item1'] = 'value'
            >>> search['item2'] = 'value'
            >>> search['item3'] = 'value'
            >>> search['item4'] = 'value'
            >>> search['item5'] = 'value'
            >>> search.insert(allow_duplicates=False)

        Instead you'll use the pattern: 
        
        ::
            >>> id_of_insert = search.Create(
            >>>             allow_duplicates=False, 
            >>>             no_overwrite_reqs=False, 
            >>>             item1='value', item2='value', 
            >>>             item3='value', item4='value', 
            >>>             item5='value')
            # The record's id is set here
            '249fabf229374715ae7e65b7061c0faf'
        

        To define a schema we set the variable `self.allrequirements`.


        ::

    """
    
    def __init__(self):
        """ Initialize the function. Pulls from existing SearchHandler

        Add `must_have` in inherited classes. Use to make certain variable names mandetory.

        Example:

        ::
            >>> def __init__(self):
            >>>    self.must_have = ["category", "name", "abbreviation"]
        


            >>> id_of_insert = search.Create(
            >>>             allow_duplicates=True, 
            >>>             no_overwrite_reqs=True)
            
            Would immediately break because `category`, `name` and `abbreviation`
            
            
        """
        super().__init__()
        self.must_have = [] # Forced fields

    def check_requirements(self, items: dict):
        """ Checks that the fields inside of `must_have` are inside of the dictioary we're going to be adding. """
        for _abs in self.must_have:
            if _abs not in items:
                raise AttributeError(
                    f"{_abs} has to be added. The absolute required variables are the following: {self.must_have}"
                )
        # """ 
        #     Insert a document. All fields defined inside of **kwargs.

        #     Parameters:
        #         allow_duplicates (bool): Determines if we want to allow duplicates of the exact same document inside of the search database.
        #         no_overwrite_must_have (bool): Determines if we're only checking for a small range of fields. Identified inside of `self.must_have`
        #         kwargs (Any): Any field we want to add to the database. It's key and value. The databse must have
        # """

    
    

    def Create(self,
               allow_duplicates=False,
               no_overwrite_must_have=False,
               **kwargs) -> str:
        """Insert a new record into redisearch.

        Args:
            allow_duplicates (bool, optional): Determines if you're going to allow for duplicates inside of the database. Defaults to False.
            no_overwrite_must_have (bool, optional): Determines if we're going to allow for more than 1 record that matches `must_have`. Defaults to False.

        Returns:
            str: The inserted record's id
        """
        self.reset()
        self.check_requirements(kwargs)

        if no_overwrite_must_have and len(self.must_have) > 0:
            
            _all = self.FForced(**kwargs)
            if len(_all) > 0:
                return _all[0].id

        for k, v in kwargs.items():
            self[k] = v

        identity = self.insert(allow_duplicates=allow_duplicates)
        return identity

    def UpdateID(self, identity: str, **kwargs):
        """ Updates a record by ID. Gives a warning if you're using a must have variable."""
        self.reset()
        for k, v in kwargs.items():
            self.replacement[k] = v
        self.update_id(identity)

    def UpdateMany(self, search_dict: dict, force_must_have=False, **replacements):
        """Replaces many records for the user. 

        Args:
            search_dict (dict): Search for what we're replacing.
            force_must_have (bool, optional): Checks that the `search_dict` has all of the `must_have` variables. Defaults to False.

        Raises:
            ValueError: If our search parameters or replacement dictionaries are empty.
        """
        self.reset()
        if not bool(search_dict) or not bool(replacements):
            raise ValueError(
                "You need to have query information AND something to replace it with."
            )
        if force_must_have:
            self.check_requirements(search_dict)

        for k, v in search_dict.items():
            self[k] = v

        for k, v in replacements.items():
            self.replacement[k] = v
        self.update()

    def Find(self, general=None,force_must=False, **fields):
        """Find Searches through the database for our records.

        Run a generalized search through the database.

        Keyword Arguments:
            general {str} -- A general term that will allow us to find terms in a fuzzy search (default: {None})
            force_must {bool} -- Checks that the search fields we have contains everything we declare as important `must_have` (default: {False})

        Raises:
            ValueError: If there's nothing for us to search field. Both fields and general are empty.

        Returns:
            [list] -- A list of descriptions. 
        """
        self.reset()
        if general is not None:
            self.general = general
        if not bool(fields):
            if general is not None:
                return self.find()
            raise ValueError("You have to search using something")
        if force_must:
            self.check_requirements(fields)

        for k, v in fields.items():
            self[k] = v
        return self.find()

    def FindById(self, identity: str):
        self.reset()
        remainder = self.pick(identity)
        return remainder
    
    def FindForced(self, **kwargs):
        self.reset()
        self.check_requirements(kwargs)
        for k in self.must_have:
            self[k] = kwargs.get(k)
        _all = self.find()
        return _all
    
    def FForced(self, **kwargs):
        for k in self.must_have:
            self[k] = kwargs.get(k)
        _all = self.find()
        return _all

    def Remove(self, **kwargs):
        self.reset()
        for k, v in kwargs.items():
            self[k] = v
        self.remove()